# Obtain shared config values
import sys
import os
from os.path import abspath, join, dirname

sys.path.append(abspath(join(dirname(__file__), "..")))
from shared_conf import *

# Releases changelog extension
extensions.append("releases")
releases_github_path = "pyinvoke/invoke"

# Default is 'local' building, but reference the public docs site when building
# under RTD.
target = join(dirname(__file__), "..", "docs", "_build")
if os.environ.get("READTHEDOCS") == "True":
    target = "https://docs.pyinvoke.org/en/latest/"
intersphinx_mapping["docs"] = (target, None)

# Sister-site links to documentation
html_theme_options["extra_nav_links"] = {
    "Documentation": "https://docs.pyinvoke.org"
}



# Obtain shared config values
import os, sys

sys.path.append(os.path.abspath(".."))
sys.path.append(os.path.abspath("../.."))
from shared_conf import *

# Enable autodoc, intersphinx
extensions.extend(["sphinx.ext.autodoc"])

# Autodoc settings
autodoc_default_options = {
    "members": True,
    "special-members": True,
}

# Sister-site links to WWW
html_theme_options["extra_nav_links"] = {
    "Main website": "https://www.pyinvoke.org"
}



import os
import sys
from io import BytesIO
from functools import wraps

try:
    import termios
except ImportError:
    # Not available on Windows
    termios = None
from contextlib import contextmanager

from unittest.mock import patch, Mock
from pytest import skip
from pytest_relaxed import trap

from invoke import Program, Runner
from invoke.terminals import WINDOWS


support = os.path.join(os.path.dirname(__file__), "_support")
ROOT = os.path.abspath(os.path.sep)


def skip_if_windows(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        if WINDOWS:
            skip()
        return fn(*args, **kwargs)

    return wrapper


@contextmanager
def support_path():
    sys.path.insert(0, support)
    try:
        yield
    finally:
        sys.path.pop(0)


def load(name):
    with support_path():
        imported = __import__(name)
        return imported


def support_file(subpath):
    with open(os.path.join(support, subpath)) as fd:
        return fd.read()


@trap
def run(invocation, program=None, invoke=True):
    """
    Run ``invocation`` via ``program``, returning output stream captures.

    ``program`` defaults to ``Program()``.

    To skip automatically assuming the argv under test starts with ``"invoke
    "``, say ``invoke=False``.

    :returns: Two-tuple of ``stdout, stderr`` strings.
    """
    if program is None:
        program = Program()
    if invoke:
        invocation = "invoke {}".format(invocation)
    program.run(invocation, exit=False)
    return sys.stdout.getvalue(), sys.stderr.getvalue()


def expect(
    invocation, out=None, err=None, program=None, invoke=True, test=None
):
    """
    Run ``invocation`` via ``program`` and expect resulting output to match.

    May give one or both of ``out``/``err`` (but not neither).

    ``program`` defaults to ``Program()``.

    To skip automatically assuming the argv under test starts with ``"invoke
    "``, say ``invoke=False``.

    To customize the operator used for testing (default: equality), use
    ``test`` (which should be an assertion wrapper of some kind).
    """
    stdout, stderr = run(invocation, program, invoke)
    # Perform tests
    if out is not None:
        if test:
            test(stdout, out)
        else:
            assert out == stdout
    if err is not None:
        if test:
            test(stderr, err)
        else:
            assert err == stderr
    # Guard against silent failures; since we say exit=False this is the only
    # real way to tell if stuff died in a manner we didn't expect.
    elif stderr:
        assert False, "Unexpected stderr: {}".format(stderr)
    return stdout, stderr


class MockSubprocess(object):
    def __init__(self, out="", err="", exit=0, isatty=None, autostart=True):
        self.out_file = BytesIO(out.encode())
        self.err_file = BytesIO(err.encode())
        self.exit = exit
        self.isatty = isatty
        if autostart:
            self.start()

    def start(self):
        # Start patchin'
        self.popen = patch("invoke.runners.Popen")
        Popen = self.popen.start()
        self.read = patch("os.read")
        read = self.read.start()
        self.sys_stdin = patch("sys.stdin", new_callable=BytesIO)
        sys_stdin = self.sys_stdin.start()
        # Setup mocks
        process = Popen.return_value
        process.returncode = self.exit
        process.stdout.fileno.return_value = 1
        process.stderr.fileno.return_value = 2
        # If requested, mock isatty to fake out pty detection
        if self.isatty is not None:
            sys_stdin.isatty = Mock(return_value=self.isatty)

        def fakeread(fileno, count):
            fd = {1: self.out_file, 2: self.err_file}[fileno]
            return fd.read(count)

        read.side_effect = fakeread
        # Return the Popen mock as it's sometimes wanted inside tests
        return Popen

    def stop(self):
        self.popen.stop()
        self.read.stop()
        self.sys_stdin.stop()


def mock_subprocess(out="", err="", exit=0, isatty=None, insert_Popen=False):
    def decorator(f):
        @wraps(f)
        # We have to include a @patch here to trick pytest into ignoring
        # the wrapped test's sometimes-there, sometimes-not mock_Popen arg. (It
        # explicitly "skips ahead" past what it perceives as patch args, even
        # though in our case those are not applying to the test function!)
        # Doesn't matter what we patch as long as it doesn't
        # actually get in our way.
        @patch("invoke.runners.pty")
        def wrapper(*args, **kwargs):
            proc = MockSubprocess(
                out=out, err=err, exit=exit, isatty=isatty, autostart=False
            )
            Popen = proc.start()
            args = list(args)
            args.pop()  # Pop the dummy patch
            if insert_Popen:
                args.append(Popen)
            try:
                f(*args, **kwargs)
            finally:
                proc.stop()

        return wrapper

    return decorator


def mock_pty(
    out="",
    err="",
    exit=0,
    isatty=None,
    trailing_error=None,
    skip_asserts=False,
    insert_os=False,
    be_childish=False,
    os_close_error=False,
):
    # Windows doesn't have ptys, so all the pty tests should be
    # skipped anyway...
    if WINDOWS:
        return skip_if_windows

    def decorator(f):
        import fcntl

        ioctl_patch = patch("invoke.runners.fcntl.ioctl", wraps=fcntl.ioctl)

        @wraps(f)
        @patch("invoke.runners.pty")
        @patch("invoke.runners.os")
        @ioctl_patch
        def wrapper(*args, **kwargs):
            args = list(args)
            pty, os, ioctl = args.pop(), args.pop(), args.pop()
            # Don't actually fork, but pretend we did (with "our" pid differing
            # depending on be_childish) & give 'parent fd' of 3 (typically,
            # first allocated non-stdin/out/err FD)
            pty.fork.return_value = (12345 if be_childish else 0), 3
            # We don't really need to care about waiting since not truly
            # forking/etc, so here we just return a nonzero "pid" + sentinel
            # wait-status value (used in some tests about WIFEXITED etc)
            os.waitpid.return_value = None, Mock(name="exitstatus")
            # Either or both of these may get called, depending...
            os.WEXITSTATUS.return_value = exit
            os.WTERMSIG.return_value = exit
            # If requested, mock isatty to fake out pty detection
            if isatty is not None:
                os.isatty.return_value = isatty
            out_file = BytesIO(out.encode())
            err_file = BytesIO(err.encode())

            def fakeread(fileno, count):
                fd = {3: out_file, 2: err_file}[fileno]
                ret = fd.read(count)
                # If asked, fake a Linux-platform trailing I/O error.
                if not ret and trailing_error:
                    raise trailing_error
                return ret

            os.read.side_effect = fakeread
            if os_close_error:
                os.close.side_effect = IOError
            if insert_os:
                args.append(os)

            # Do the thing!!!
            f(*args, **kwargs)

            # Short-circuit if we raised an error in fakeread()
            if trailing_error:
                return
            # Sanity checks to make sure the stuff we mocked, actually got ran!
            pty.fork.assert_called_with()
            # Skip rest of asserts if we pretended to be the child
            if be_childish:
                return
            # Expect a get, and then later set, of terminal window size
            assert ioctl.call_args_list[0][0][1] == termios.TIOCGWINSZ
            assert ioctl.call_args_list[1][0][1] == termios.TIOCSWINSZ
            if not skip_asserts:
                for name in ("execve", "waitpid"):
                    assert getattr(os, name).called
                # Ensure at least one of the exit status getters was called
                assert os.WEXITSTATUS.called or os.WTERMSIG.called
                # Ensure something closed the pty FD
                os.close.assert_called_once_with(3)

        return wrapper

    return decorator


class _Dummy(Runner):
    """
    Dummy runner subclass that does minimum work required to execute run().

    It also serves as a convenient basic API checker; failure to update it to
    match the current Runner API will cause TypeErrors, NotImplementedErrors,
    and similar.
    """

    # Neuter the input loop sleep, so tests aren't slow (at the expense of CPU,
    # which isn't a problem for testing).
    input_sleep = 0

    def start(self, command, shell, env, timeout=None):
        pass

    def read_proc_stdout(self, num_bytes):
        return ""

    def read_proc_stderr(self, num_bytes):
        return ""

    def _write_proc_stdin(self, data):
        pass

    def close_proc_stdin(self):
        pass

    @property
    def process_is_finished(self):
        return True

    def returncode(self):
        return 0

    @property
    def timed_out(self):
        return False


# Dummy command that will blow up if it ever truly hits a real shell.
_ = "nope"


# Runner that fakes ^C during subprocess exec
class _KeyboardInterruptingRunner(_Dummy):
    def __init__(self, *args, **kwargs):
        super(_KeyboardInterruptingRunner, self).__init__(*args, **kwargs)
        self._interrupted = False

    # Trigger KeyboardInterrupt during wait()
    def wait(self):
        if not self._interrupted:
            self._interrupted = True
            raise KeyboardInterrupt

    # But also, after that has been done, pretend subprocess shutdown happened
    # (or we will loop forever).
    def process_is_finished(self):
        return self._interrupted


class OhNoz(Exception):
    pass



from invoke.collection import Collection
from invoke.parser import Parser
from invoke.tasks import task


class CLIParsing:
    """
    High level parsing tests
    """

    def setup_method(self):
        @task(positional=[], iterable=["my_list"], incrementable=["verbose"])
        def my_task(
            c,
            mystring,
            s,
            boolean=False,
            b=False,
            v=False,
            long_name=False,
            true_bool=True,
            _leading_underscore=False,
            trailing_underscore_=False,
            my_list=None,
            verbose=0,
        ):
            pass

        @task(aliases=["my_task27"])
        def my_task2(c):
            pass

        @task(default=True)
        def my_task3(c, mystring):
            pass

        @task
        def my_task4(c, clean=False, browse=False):
            pass

        @task(aliases=["other"], default=True)
        def sub_task(c):
            pass

        sub_coll = Collection("sub_coll", sub_task)
        self.c = Collection(my_task, my_task2, my_task3, my_task4, sub_coll)

    def _parser(self):
        return Parser(self.c.to_contexts())

    def _parse(self, argstr):
        return self._parser().parse_argv(argstr.split())

    def _compare(self, invoke, flagname, value):
        invoke = "my-task " + invoke
        result = self._parse(invoke)
        assert result[0].args[flagname].value == value

    def _compare_names(self, given, real):
        assert self._parse(given)[0].name == real

    def underscored_flags_can_be_given_as_dashed(self):
        self._compare("--long-name", "long_name", True)

    def leading_underscores_are_ignored(self):
        self._compare("--leading-underscore", "_leading_underscore", True)

    def trailing_underscores_are_ignored(self):
        self._compare("--trailing-underscore", "trailing_underscore_", True)

    def inverse_boolean_flags(self):
        self._compare("--no-true-bool", "true_bool", False)

    def namespaced_task(self):
        self._compare_names("sub-coll.sub-task", "sub-coll.sub-task")

    def aliases(self):
        self._compare_names("my-task27", "my-task2")

    def subcollection_aliases(self):
        self._compare_names("sub-coll.other", "sub-coll.sub-task")

    def subcollection_default_tasks(self):
        self._compare_names("sub-coll", "sub-coll.sub-task")

    def boolean_args(self):
        "my-task --boolean"
        self._compare("--boolean", "boolean", True)

    def flag_then_space_then_value(self):
        "my-task --mystring foo"
        self._compare("--mystring foo", "mystring", "foo")

    def flag_then_equals_sign_then_value(self):
        "my-task --mystring=foo"
        self._compare("--mystring=foo", "mystring", "foo")

    def short_boolean_flag(self):
        "my-task -b"
        self._compare("-b", "b", True)

    def short_flag_then_space_then_value(self):
        "my-task -s value"
        self._compare("-s value", "s", "value")

    def short_flag_then_equals_sign_then_value(self):
        "my-task -s=value"
        self._compare("-s=value", "s", "value")

    def short_flag_with_adjacent_value(self):
        "my-task -svalue"
        r = self._parse("my-task -svalue")
        assert r[0].args.s.value == "value"

    def _flag_value_task(self, value):
        r = self._parse("my-task -s {} my-task2".format(value))
        assert len(r) == 2
        assert r[0].name == "my-task"
        assert r[0].args.s.value == value
        assert r[1].name == "my-task2"

    def flag_value_then_task(self):
        "my-task -s value my-task2"
        self._flag_value_task("value")

    def flag_value_same_as_task_name(self):
        "my-task -s my-task2 my-task2"
        self._flag_value_task("my-task2")

    def three_tasks_with_args(self):
        "my-task --boolean my-task3 --mystring foo my-task2"
        r = self._parse("my-task --boolean my-task3 --mystring foo my-task2")
        assert len(r) == 3
        assert [x.name for x in r] == ["my-task", "my-task3", "my-task2"]
        assert r[0].args.boolean.value
        assert r[1].args.mystring.value == "foo"

    def tasks_with_duplicately_named_kwargs(self):
        "my-task --mystring foo my-task3 --mystring bar"
        r = self._parse("my-task --mystring foo my-task3 --mystring bar")
        assert r[0].name == "my-task"
        assert r[0].args.mystring.value == "foo"
        assert r[1].name == "my-task3"
        assert r[1].args.mystring.value == "bar"

    def multiple_short_flags_adjacent(self):
        "my-task -bv (and inverse)"
        for args in ("-bv", "-vb"):
            r = self._parse("my-task {}".format(args))
            a = r[0].args
            assert a.b.value
            assert a.v.value

    def list_type_flag_can_be_given_N_times_building_a_list(self):
        "my-task --my-list foo --my-list bar"
        # Test both the singular and plural cases, just to be safe.
        self._compare("--my-list foo", "my-list", ["foo"])
        self._compare("--my-list foo --my-list bar", "my-list", ["foo", "bar"])

    def incrementable_type_flag_can_be_used_as_a_switch_or_counter(self):
        "my-task -v, -vv, -vvvvv etc, except with explicit --verbose"
        self._compare("", "verbose", 0)
        self._compare("--verbose", "verbose", 1)
        self._compare("--verbose --verbose --verbose", "verbose", 3)



from invoke.tasks import task


@task(aliases=("bar", "otherbar"))
def foo(c):
    """
    Foo the bar.
    """
    pass


@task
def foo2(c):
    """
    Foo the bar:

      example code

    Added in 1.0
    """
    pass


@task
def foo3(c):
    """Foo the other bar:

      example code

    Added in 1.1
    """
    pass


@task(default=True)
def biz(c):
    pass


@task(help={"why": "Motive", "who": "Who to punch"})
def punch(c, who, why):
    pass


@task(positional=["pos"])
def one_positional(c, pos, nonpos):
    pass


@task(positional=["pos1", "pos2"])
def two_positionals(c, pos1, pos2, nonpos):
    pass


@task
def implicit_positionals(c, pos1, pos2, nonpos=None):
    pass


@task(optional=["myopt"])
def optional_values(c, myopt):
    pass


@task(iterable=["mylist"])
def iterable_values(c, mylist=None):
    pass


@task(incrementable=["verbose"])
def incrementable_values(c, verbose=None):
    pass



import pickle
import os
from os.path import join

from unittest.mock import patch, call, Mock
import pytest
from pytest_relaxed import raises

from invoke import config as config_mod  # for accessing mocks
from invoke.runners import Local
from invoke.config import Config
from invoke.exceptions import (
    AmbiguousEnvVar,
    UncastableEnvVar,
    UnknownFileType,
    UnpicklableConfigMember,
)

from _util import skip_if_windows, support


pytestmark = pytest.mark.usefixtures("integration")


CONFIGS_PATH = "configs"
TYPES = ("yaml", "yml", "json", "python")


def _load(kwarg, type_, **kwargs):
    path = join(CONFIGS_PATH, type_ + "/")
    kwargs[kwarg] = path
    return Config(**kwargs)


class Config_:
    class class_attrs:
        # TODO: move all other non-data-bearing kwargs to this mode
        class prefix:
            def defaults_to_invoke(self):
                assert Config().prefix == "invoke"

            @patch.object(Config, "_load_yaml")
            def informs_config_filenames(self, load_yaml):
                class MyConf(Config):
                    prefix = "other"

                MyConf(system_prefix="dir/")
                load_yaml.assert_any_call("dir/other.yaml")

            def informs_env_var_prefix(self):
                os.environ["OTHER_FOO"] = "bar"

                class MyConf(Config):
                    prefix = "other"

                c = MyConf(defaults={"foo": "notbar"})
                c.load_shell_env()
                assert c.foo == "bar"

        class file_prefix:
            def defaults_to_None(self):
                assert Config().file_prefix is None

            @patch.object(Config, "_load_yaml")
            def informs_config_filenames(self, load_yaml):
                class MyConf(Config):
                    file_prefix = "other"

                MyConf(system_prefix="dir/")
                load_yaml.assert_any_call("dir/other.yaml")

        class env_prefix:
            def defaults_to_None(self):
                assert Config().env_prefix is None

            def informs_env_vars_loaded(self):
                os.environ["OTHER_FOO"] = "bar"

                class MyConf(Config):
                    env_prefix = "other"

                c = MyConf(defaults={"foo": "notbar"})
                c.load_shell_env()
                assert c.foo == "bar"

    class global_defaults:
        @skip_if_windows
        def basic_settings(self):
            # Just a catchall for what the baseline config settings should
            # be...for some reason we're not actually capturing all of these
            # reliably (even if their defaults are often implied by the tests
            # which override them, e.g. runner tests around warn=True, etc).
            expected = {
                "run": {
                    "asynchronous": False,
                    "disown": False,
                    "dry": False,
                    "echo": False,
                    "echo_format": "\033[1;37m{command}\033[0m",
                    "echo_stdin": None,
                    "encoding": None,
                    "env": {},
                    "err_stream": None,
                    "fallback": True,
                    "hide": None,
                    "in_stream": None,
                    "out_stream": None,
                    "pty": False,
                    "replace_env": False,
                    "shell": "/bin/bash",
                    "warn": False,
                    "watchers": [],
                },
                "runners": {"local": Local},
                "sudo": {
                    "password": None,
                    "prompt": "[sudo] password: ",
                    "user": None,
                },
                "tasks": {
                    "auto_dash_names": True,
                    "collection_name": "tasks",
                    "dedupe": True,
                    "executor_class": None,
                    "ignore_unknown_help": False,
                    "search_root": None,
                },
                "timeouts": {"command": None},
            }
            assert Config.global_defaults() == expected

    class init:
        "__init__"

        def can_be_empty(self):
            assert Config().__class__ == Config  # derp

        @patch.object(Config, "_load_yaml")
        def configure_global_location_prefix(self, load_yaml):
            # This is a bit funky but more useful than just replicating the
            # same test farther down?
            Config(system_prefix="meh/")
            load_yaml.assert_any_call("meh/invoke.yaml")

        @skip_if_windows
        @patch.object(Config, "_load_yaml")
        def default_system_prefix_is_etc(self, load_yaml):
            # TODO: make this work on Windows somehow without being a total
            # tautology? heh.
            Config()
            load_yaml.assert_any_call("/etc/invoke.yaml")

        @patch.object(Config, "_load_yaml")
        def configure_user_location_prefix(self, load_yaml):
            Config(user_prefix="whatever/")
            load_yaml.assert_any_call("whatever/invoke.yaml")

        @patch.object(Config, "_load_yaml")
        def default_user_prefix_is_homedir_plus_dot(self, load_yaml):
            Config()
            # NOTE: expects autouse fixture which patches expanduser
            # Make sure we called expanduser() with tilde syntax
            config_mod.expanduser.assert_any_call("~/.invoke.yaml")
            # Make sure result of that call was passed into load_yaml
            load_yaml.assert_any_call(config_mod.expanduser("~/.invoke.yaml"))

        @patch.object(Config, "_load_yaml")
        def configure_project_location(self, load_yaml):
            Config(project_location="someproject").load_project()
            load_yaml.assert_any_call(join("someproject", "invoke.yaml"))

        @patch.object(Config, "_load_yaml")
        def configure_runtime_path(self, load_yaml):
            Config(runtime_path="some/path.yaml").load_runtime()
            load_yaml.assert_any_call("some/path.yaml")

        def accepts_defaults_dict_kwarg(self):
            c = Config(defaults={"super": "low level"})
            assert c.super == "low level"

        def overrides_dict_is_first_posarg(self):
            c = Config({"new": "data", "run": {"hide": True}})
            assert c.run.hide is True  # default is False
            assert c.run.warn is False  # in global defaults, untouched
            assert c.new == "data"  # data only present at overrides layer

        def overrides_dict_is_also_a_kwarg(self):
            c = Config(overrides={"run": {"hide": True}})
            assert c.run.hide is True

        @patch.object(Config, "load_system")
        @patch.object(Config, "load_user")
        @patch.object(Config, "merge")
        def system_and_user_files_loaded_automatically(
            self, merge, load_u, load_s
        ):
            Config()
            load_s.assert_called_once_with(merge=False)
            load_u.assert_called_once_with(merge=False)
            merge.assert_called_once_with()

        @patch.object(Config, "load_system")
        @patch.object(Config, "load_user")
        def can_defer_loading_system_and_user_files(self, load_u, load_s):
            config = Config(lazy=True)
            assert not load_s.called
            assert not load_u.called
            # Make sure default levels are still in place! (When bug present,
            # i.e. merge() never called, config appears effectively empty.)
            assert config.run.echo is False

    class basic_API:
        "Basic API components"

        def can_be_used_directly_after_init(self):
            # No load() here...
            c = Config({"lots of these": "tests look similar"})
            assert c["lots of these"] == "tests look similar"

        def allows_dict_and_attr_access(self):
            # TODO: combine with tests for Context probably
            c = Config({"foo": "bar"})
            assert c.foo == "bar"
            assert c["foo"] == "bar"

        def nested_dict_values_also_allow_dual_access(self):
            # TODO: ditto
            c = Config({"foo": "bar", "biz": {"baz": "boz"}})
            # Sanity check - nested doesn't somehow kill simple top level
            assert c.foo == "bar"
            assert c["foo"] == "bar"
            # Actual check
            assert c.biz.baz == "boz"
            assert c["biz"]["baz"] == "boz"
            assert c.biz["baz"] == "boz"
            assert c["biz"].baz == "boz"

        def attr_access_has_useful_error_msg(self):
            c = Config()
            try:
                c.nope
            except AttributeError as e:
                expected = """
No attribute or config key found for 'nope'

Valid keys: ['run', 'runners', 'sudo', 'tasks', 'timeouts']

Valid real attributes: ['clear', 'clone', 'env_prefix', 'file_prefix', 'from_data', 'global_defaults', 'load_base_conf_files', 'load_collection', 'load_defaults', 'load_overrides', 'load_project', 'load_runtime', 'load_shell_env', 'load_system', 'load_user', 'merge', 'pop', 'popitem', 'prefix', 'set_project_location', 'set_runtime_path', 'setdefault', 'update']
""".strip()  # noqa
                assert str(e) == expected
            else:
                assert False, "Didn't get an AttributeError on bad key!"

        def subkeys_get_merged_not_overwritten(self):
            # Ensures nested keys merge deeply instead of shallowly.
            defaults = {"foo": {"bar": "baz"}}
            overrides = {"foo": {"notbar": "notbaz"}}
            c = Config(defaults=defaults, overrides=overrides)
            assert c.foo.notbar == "notbaz"
            assert c.foo.bar == "baz"

        def is_iterable_like_dict(self):
            c = Config(defaults={"a": 1, "b": 2})
            assert set(c.keys()) == {"a", "b"}
            assert set(list(c)) == {"a", "b"}

        def supports_readonly_dict_protocols(self):
            # Use single-keypair dict to avoid sorting problems in tests.
            c = Config(defaults={"foo": "bar"})
            c2 = Config(defaults={"foo": "bar"})
            assert "foo" in c
            assert "foo" in c2  # mostly just to trigger loading :x
            assert c == c2
            assert len(c) == 1
            assert c.get("foo") == "bar"
            assert list(c.items()) == [("foo", "bar")]
            assert list(c.keys()) == ["foo"]
            assert list(c.values()) == ["bar"]

        class runtime_loading_of_defaults_and_overrides:
            def defaults_can_be_given_via_method(self):
                c = Config()
                assert "foo" not in c
                c.load_defaults({"foo": "bar"})
                assert c.foo == "bar"

            def defaults_can_skip_merging(self):
                c = Config()
                c.load_defaults({"foo": "bar"}, merge=False)
                assert "foo" not in c
                c.merge()
                assert c.foo == "bar"

            def overrides_can_be_given_via_method(self):
                c = Config(defaults={"foo": "bar"})
                assert c.foo == "bar"  # defaults level
                c.load_overrides({"foo": "notbar"})
                assert c.foo == "notbar"  # overrides level

            def overrides_can_skip_merging(self):
                c = Config()
                c.load_overrides({"foo": "bar"}, merge=False)
                assert "foo" not in c
                c.merge()
                assert c.foo == "bar"

        class deletion_methods:
            def pop(self):
                # Root
                c = Config(defaults={"foo": "bar"})
                assert c.pop("foo") == "bar"
                assert c == {}
                # With the default arg
                assert c.pop("wut", "fine then") == "fine then"
                # Leaf (different key to avoid AmbiguousMergeError)
                c.nested = {"leafkey": "leafval"}
                assert c.nested.pop("leafkey") == "leafval"
                assert c == {"nested": {}}

            def delitem(self):
                "__delitem__"
                c = Config(defaults={"foo": "bar"})
                del c["foo"]
                assert c == {}
                c.nested = {"leafkey": "leafval"}
                del c.nested["leafkey"]
                assert c == {"nested": {}}

            def delattr(self):
                "__delattr__"
                c = Config(defaults={"foo": "bar"})
                del c.foo
                assert c == {}
                c.nested = {"leafkey": "leafval"}
                del c.nested.leafkey
                assert c == {"nested": {}}

            def clear(self):
                c = Config(defaults={"foo": "bar"})
                c.clear()
                assert c == {}
                c.nested = {"leafkey": "leafval"}
                c.nested.clear()
                assert c == {"nested": {}}

            def popitem(self):
                c = Config(defaults={"foo": "bar"})
                assert c.popitem() == ("foo", "bar")
                assert c == {}
                c.nested = {"leafkey": "leafval"}
                assert c.nested.popitem() == ("leafkey", "leafval")
                assert c == {"nested": {}}

        class modification_methods:
            def setitem(self):
                c = Config(defaults={"foo": "bar"})
                c["foo"] = "notbar"
                assert c.foo == "notbar"
                del c["foo"]
                c["nested"] = {"leafkey": "leafval"}
                assert c == {"nested": {"leafkey": "leafval"}}

            def setdefault(self):
                c = Config({"foo": "bar", "nested": {"leafkey": "leafval"}})
                assert c.setdefault("foo") == "bar"
                assert c.nested.setdefault("leafkey") == "leafval"
                assert c.setdefault("notfoo", "notbar") == "notbar"
                assert c.notfoo == "notbar"
                nested = c.nested.setdefault("otherleaf", "otherval")
                assert nested == "otherval"
                assert c.nested.otherleaf == "otherval"

            def update(self):
                c = Config(
                    defaults={"foo": "bar", "nested": {"leafkey": "leafval"}}
                )
                # Regular update(dict)
                c.update({"foo": "notbar"})
                assert c.foo == "notbar"
                c.nested.update({"leafkey": "otherval"})
                assert c.nested.leafkey == "otherval"
                # Apparently allowed but wholly useless
                c.update()
                expected = {"foo": "notbar", "nested": {"leafkey": "otherval"}}
                assert c == expected
                # Kwarg edition
                c.update(foo="otherbar")
                assert c.foo == "otherbar"
                # Iterator of 2-tuples edition
                c.nested.update(
                    [("leafkey", "yetanotherval"), ("newleaf", "turnt")]
                )
                assert c.nested.leafkey == "yetanotherval"
                assert c.nested.newleaf == "turnt"

        def reinstatement_of_deleted_values_works_ok(self):
            # Sounds like a stupid thing to test, but when we have to track
            # deletions and mutations manually...it's an easy thing to overlook
            c = Config(defaults={"foo": "bar"})
            assert c.foo == "bar"
            del c["foo"]
            # Sanity checks
            assert "foo" not in c
            assert len(c) == 0
            # Put it back again...as a different value, for funsies
            c.foo = "formerly bar"
            # And make sure it stuck
            assert c.foo == "formerly bar"

        def deleting_parent_keys_of_deleted_keys_subsumes_them(self):
            c = Config({"foo": {"bar": "biz"}})
            del c.foo["bar"]
            del c.foo
            # Make sure we didn't somehow still end up with {'foo': {'bar':
            # None}}
            assert c._deletions == {"foo": None}

        def supports_mutation_via_attribute_access(self):
            c = Config({"foo": "bar"})
            assert c.foo == "bar"
            c.foo = "notbar"
            assert c.foo == "notbar"
            assert c["foo"] == "notbar"

        def supports_nested_mutation_via_attribute_access(self):
            c = Config({"foo": {"bar": "biz"}})
            assert c.foo.bar == "biz"
            c.foo.bar = "notbiz"
            assert c.foo.bar == "notbiz"
            assert c["foo"]["bar"] == "notbiz"

        def real_attrs_and_methods_win_over_attr_proxying(self):
            # Setup
            class MyConfig(Config):
                myattr = None

                def mymethod(self):
                    return 7

            c = MyConfig({"myattr": "foo", "mymethod": "bar"})
            # By default, attr and config value separate
            assert c.myattr is None
            assert c["myattr"] == "foo"
            # After a setattr, same holds true
            c.myattr = "notfoo"
            assert c.myattr == "notfoo"
            assert c["myattr"] == "foo"
            # Method and config value separate
            assert callable(c.mymethod)
            assert c.mymethod() == 7
            assert c["mymethod"] == "bar"
            # And same after setattr
            def monkeys():
                return 13

            c.mymethod = monkeys
            assert c.mymethod() == 13
            assert c["mymethod"] == "bar"

        def config_itself_stored_as_private_name(self):
            # I.e. one can refer to a key called 'config', which is relatively
            # commonplace (e.g. <Config>.myservice.config -> a config file
            # contents or path or etc)
            c = Config()
            c["foo"] = {"bar": "baz"}
            c["whatever"] = {"config": "myconfig"}
            assert c.foo.bar == "baz"
            assert c.whatever.config == "myconfig"

        def inherited_real_attrs_also_win_over_config_keys(self):
            class MyConfigParent(Config):
                parent_attr = 17

            class MyConfig(MyConfigParent):
                pass

            c = MyConfig()
            assert c.parent_attr == 17
            c.parent_attr = 33
            oops = "Oops! Looks like config won over real attr!"
            assert "parent_attr" not in c, oops
            assert c.parent_attr == 33
            c["parent_attr"] = "fifteen"
            assert c.parent_attr == 33
            assert c["parent_attr"] == "fifteen"

        def nonexistent_attrs_can_be_set_to_create_new_top_level_configs(self):
            # I.e. some_config.foo = 'bar' is like some_config['foo'] = 'bar'.
            # When this test breaks it usually means some_config.foo = 'bar'
            # sets a regular attribute - and the configuration itself is never
            # touched!
            c = Config()
            c.some_setting = "some_value"
            assert c["some_setting"] == "some_value"

        def nonexistent_attr_setting_works_nested_too(self):
            c = Config()
            c.a_nest = {}
            assert c["a_nest"] == {}
            c.a_nest.an_egg = True
            assert c["a_nest"]["an_egg"]

        def string_display(self):
            "__str__ and friends"
            config = Config(defaults={"foo": "bar"})
            assert repr(config) == "<Config: {'foo': 'bar'}>"

        def merging_does_not_wipe_user_modifications_or_deletions(self):
            c = Config({"foo": {"bar": "biz"}, "error": True})
            c.foo.bar = "notbiz"
            del c["error"]
            assert c["foo"]["bar"] == "notbiz"
            assert "error" not in c
            c.merge()
            # Will be back to 'biz' if user changes don't get saved on their
            # own (previously they are just mutations on the cached central
            # config)
            assert c["foo"]["bar"] == "notbiz"
            # And this would still be here, too
            assert "error" not in c

    class config_file_loading:
        "Configuration file loading"

        def system_global(self):
            "Systemwide conf files"
            # NOTE: using lazy=True to avoid autoloading so we can prove
            # load_system() works.
            for type_ in TYPES:
                config = _load("system_prefix", type_, lazy=True)
                assert "outer" not in config
                config.load_system()
                assert config.outer.inner.hooray == type_

        def system_can_skip_merging(self):
            config = _load("system_prefix", "yml", lazy=True)
            assert "outer" not in config._system
            assert "outer" not in config
            config.load_system(merge=False)
            # Test that we loaded into the per-level dict, but not the
            # central/merged config.
            assert "outer" in config._system
            assert "outer" not in config

        def user_specific(self):
            "User-specific conf files"
            # NOTE: using lazy=True to avoid autoloading so we can prove
            # load_user() works.
            for type_ in TYPES:
                config = _load("user_prefix", type_, lazy=True)
                assert "outer" not in config
                config.load_user()
                assert config.outer.inner.hooray == type_

        def user_can_skip_merging(self):
            config = _load("user_prefix", "yml", lazy=True)
            assert "outer" not in config._user
            assert "outer" not in config
            config.load_user(merge=False)
            # Test that we loaded into the per-level dict, but not the
            # central/merged config.
            assert "outer" in config._user
            assert "outer" not in config

        def project_specific(self):
            "Local-to-project conf files"
            for type_ in TYPES:
                c = Config(project_location=join(CONFIGS_PATH, type_))
                assert "outer" not in c
                c.load_project()
                assert c.outer.inner.hooray == type_

        def project_can_skip_merging(self):
            config = Config(
                project_location=join(CONFIGS_PATH, "yml"), lazy=True
            )
            assert "outer" not in config._project
            assert "outer" not in config
            config.load_project(merge=False)
            # Test that we loaded into the per-level dict, but not the
            # central/merged config.
            assert "outer" in config._project
            assert "outer" not in config

        def loads_no_project_specific_file_if_no_project_location_given(self):
            c = Config()
            assert c._project_path is None
            c.load_project()
            assert list(c._project.keys()) == []
            defaults = ["tasks", "run", "runners", "sudo", "timeouts"]
            assert set(c.keys()) == set(defaults)

        def project_location_can_be_set_after_init(self):
            c = Config()
            assert "outer" not in c
            c.set_project_location(join(CONFIGS_PATH, "yml"))
            c.load_project()
            assert c.outer.inner.hooray == "yml"

        def runtime_conf_via_cli_flag(self):
            c = Config(runtime_path=join(CONFIGS_PATH, "yaml", "invoke.yaml"))
            c.load_runtime()
            assert c.outer.inner.hooray == "yaml"

        def runtime_can_skip_merging(self):
            path = join(CONFIGS_PATH, "yaml", "invoke.yaml")
            config = Config(runtime_path=path, lazy=True)
            assert "outer" not in config._runtime
            assert "outer" not in config
            config.load_runtime(merge=False)
            # Test that we loaded into the per-level dict, but not the
            # central/merged config.
            assert "outer" in config._runtime
            assert "outer" not in config

        @raises(UnknownFileType)
        def unknown_suffix_in_runtime_path_raises_useful_error(self):
            c = Config(runtime_path=join(CONFIGS_PATH, "screw.ini"))
            c.load_runtime()

        def python_modules_dont_load_special_vars(self):
            "Python modules don't load special vars"
            # Borrow another test's Python module.
            c = _load("system_prefix", "python")
            # Sanity test that lowercase works
            assert c.outer.inner.hooray == "python"
            # Real test that builtins, etc are stripped out
            for special in ("builtins", "file", "package", "name", "doc"):
                assert "__{}__".format(special) not in c

        def python_modules_except_usefully_on_unpicklable_modules(self):
            # Re: #556; when bug present, a TypeError pops up instead (granted,
            # at merge time, but we want it to raise ASAP, so we're testing the
            # intended new behavior: raising at config load time.
            c = Config()
            c.set_runtime_path(join(support, "has_modules.py"))
            expected = r"'os' is a module.*giving a tasks file.*mistake"
            with pytest.raises(UnpicklableConfigMember, match=expected):
                c.load_runtime(merge=False)

        @patch("invoke.config.debug")
        def nonexistent_files_are_skipped_and_logged(self, mock_debug):
            c = Config()
            c._load_yml = Mock(side_effect=IOError(2, "aw nuts"))
            c.set_runtime_path("is-a.yml")  # Triggers use of _load_yml
            c.load_runtime()
            mock_debug.assert_any_call("Didn't see any is-a.yml, skipping.")

        @raises(IOError)
        def non_missing_file_IOErrors_are_raised(self):
            c = Config()
            c._load_yml = Mock(side_effect=IOError(17, "uh, what?"))
            c.set_runtime_path("is-a.yml")  # Triggers use of _load_yml
            c.load_runtime()

    class collection_level_config_loading:
        def performed_explicitly_and_directly(self):
            # TODO: do we want to update the other levels to allow 'direct'
            # loading like this, now that they all have explicit methods?
            c = Config()
            assert "foo" not in c
            c.load_collection({"foo": "bar"})
            assert c.foo == "bar"

        def merging_can_be_deferred(self):
            c = Config()
            assert "foo" not in c._collection
            assert "foo" not in c
            c.load_collection({"foo": "bar"}, merge=False)
            assert "foo" in c._collection
            assert "foo" not in c

    class comparison_and_hashing:
        def comparison_looks_at_merged_config(self):
            c1 = Config(defaults={"foo": {"bar": "biz"}})
            # Empty defaults to suppress global_defaults
            c2 = Config(defaults={}, overrides={"foo": {"bar": "biz"}})
            assert c1 is not c2
            assert c1._defaults != c2._defaults
            assert c1 == c2

        def allows_comparison_with_real_dicts(self):
            c = Config({"foo": {"bar": "biz"}})
            assert c["foo"] == {"bar": "biz"}

        @raises(TypeError)
        def is_explicitly_not_hashable(self):
            hash(Config())

    class env_vars:
        "Environment variables"

        def base_case_defaults_to_INVOKE_prefix(self):
            os.environ["INVOKE_FOO"] = "bar"
            c = Config(defaults={"foo": "notbar"})
            c.load_shell_env()
            assert c.foo == "bar"

        def non_predeclared_settings_do_not_get_consumed(self):
            os.environ["INVOKE_HELLO"] = "is it me you're looking for?"
            c = Config()
            c.load_shell_env()
            assert "HELLO" not in c
            assert "hello" not in c

        def underscores_top_level(self):
            os.environ["INVOKE_FOO_BAR"] = "biz"
            c = Config(defaults={"foo_bar": "notbiz"})
            c.load_shell_env()
            assert c.foo_bar == "biz"

        def underscores_nested(self):
            os.environ["INVOKE_FOO_BAR"] = "biz"
            c = Config(defaults={"foo": {"bar": "notbiz"}})
            c.load_shell_env()
            assert c.foo.bar == "biz"

        def both_types_of_underscores_mixed(self):
            os.environ["INVOKE_FOO_BAR_BIZ"] = "baz"
            c = Config(defaults={"foo_bar": {"biz": "notbaz"}})
            c.load_shell_env()
            assert c.foo_bar.biz == "baz"

        @raises(AmbiguousEnvVar)
        def ambiguous_underscores_dont_guess(self):
            os.environ["INVOKE_FOO_BAR"] = "biz"
            c = Config(defaults={"foo_bar": "wat", "foo": {"bar": "huh"}})
            c.load_shell_env()

        class type_casting:
            def strings_replaced_with_env_value(self):
                os.environ["INVOKE_FOO"] = "myvalue"
                c = Config(defaults={"foo": "myoldvalue"})
                c.load_shell_env()
                assert c.foo == "myvalue"
                assert isinstance(c.foo, str)

            def None_replaced(self):
                os.environ["INVOKE_FOO"] = "something"
                c = Config(defaults={"foo": None})
                c.load_shell_env()
                assert c.foo == "something"

            def booleans(self):
                for input_, result in (
                    ("0", False),
                    ("1", True),
                    ("", False),
                    ("meh", True),
                    ("false", True),
                ):
                    os.environ["INVOKE_FOO"] = input_
                    c = Config(defaults={"foo": bool()})
                    c.load_shell_env()
                    assert c.foo == result

            def boolean_type_inputs_with_non_boolean_defaults(self):
                for input_ in ("0", "1", "", "meh", "false"):
                    os.environ["INVOKE_FOO"] = input_
                    c = Config(defaults={"foo": "bar"})
                    c.load_shell_env()
                    assert c.foo == input_

            def numeric_types_become_casted(self):
                tests = [
                    (int, "5", 5),
                    (float, "5.5", 5.5),
                ]
                for old, new_, result in tests:
                    os.environ["INVOKE_FOO"] = new_
                    c = Config(defaults={"foo": old()})
                    c.load_shell_env()
                    assert c.foo == result

            def arbitrary_types_work_too(self):
                os.environ["INVOKE_FOO"] = "whatever"

                class Meh(object):
                    def __init__(self, thing=None):
                        pass

                old_obj = Meh()
                c = Config(defaults={"foo": old_obj})
                c.load_shell_env()
                assert isinstance(c.foo, Meh)
                assert c.foo is not old_obj

            class uncastable_types:
                @raises(UncastableEnvVar)
                def _uncastable_type(self, default):
                    os.environ["INVOKE_FOO"] = "stuff"
                    c = Config(defaults={"foo": default})
                    c.load_shell_env()

                def lists(self):
                    self._uncastable_type(["a", "list"])

                def tuples(self):
                    self._uncastable_type(("a", "tuple"))

    class hierarchy:
        "Config hierarchy in effect"

        #
        # NOTE: most of these just leverage existing test fixtures (which live
        # in their own directories & have differing values for the 'hooray'
        # key), since we normally don't need more than 2-3 different file
        # locations for any one test.
        #

        def collection_overrides_defaults(self):
            c = Config(defaults={"nested": {"setting": "default"}})
            c.load_collection({"nested": {"setting": "collection"}})
            assert c.nested.setting == "collection"

        def systemwide_overrides_collection(self):
            c = Config(system_prefix=join(CONFIGS_PATH, "yaml/"))
            c.load_collection({"outer": {"inner": {"hooray": "defaults"}}})
            assert c.outer.inner.hooray == "yaml"

        def user_overrides_systemwide(self):
            c = Config(
                system_prefix=join(CONFIGS_PATH, "yaml/"),
                user_prefix=join(CONFIGS_PATH, "json/"),
            )
            assert c.outer.inner.hooray == "json"

        def user_overrides_collection(self):
            c = Config(user_prefix=join(CONFIGS_PATH, "json/"))
            c.load_collection({"outer": {"inner": {"hooray": "defaults"}}})
            assert c.outer.inner.hooray == "json"

        def project_overrides_user(self):
            c = Config(
                user_prefix=join(CONFIGS_PATH, "json/"),
                project_location=join(CONFIGS_PATH, "yaml"),
            )
            c.load_project()
            assert c.outer.inner.hooray == "yaml"

        def project_overrides_systemwide(self):
            c = Config(
                system_prefix=join(CONFIGS_PATH, "json/"),
                project_location=join(CONFIGS_PATH, "yaml"),
            )
            c.load_project()
            assert c.outer.inner.hooray == "yaml"

        def project_overrides_collection(self):
            c = Config(project_location=join(CONFIGS_PATH, "yaml"))
            c.load_project()
            c.load_collection({"outer": {"inner": {"hooray": "defaults"}}})
            assert c.outer.inner.hooray == "yaml"

        def env_vars_override_project(self):
            os.environ["INVOKE_OUTER_INNER_HOORAY"] = "env"
            c = Config(project_location=join(CONFIGS_PATH, "yaml"))
            c.load_project()
            c.load_shell_env()
            assert c.outer.inner.hooray == "env"

        def env_vars_override_user(self):
            os.environ["INVOKE_OUTER_INNER_HOORAY"] = "env"
            c = Config(user_prefix=join(CONFIGS_PATH, "yaml/"))
            c.load_shell_env()
            assert c.outer.inner.hooray == "env"

        def env_vars_override_systemwide(self):
            os.environ["INVOKE_OUTER_INNER_HOORAY"] = "env"
            c = Config(system_prefix=join(CONFIGS_PATH, "yaml/"))
            c.load_shell_env()
            assert c.outer.inner.hooray == "env"

        def env_vars_override_collection(self):
            os.environ["INVOKE_OUTER_INNER_HOORAY"] = "env"
            c = Config()
            c.load_collection({"outer": {"inner": {"hooray": "defaults"}}})
            c.load_shell_env()
            assert c.outer.inner.hooray == "env"

        def runtime_overrides_env_vars(self):
            os.environ["INVOKE_OUTER_INNER_HOORAY"] = "env"
            c = Config(runtime_path=join(CONFIGS_PATH, "json", "invoke.json"))
            c.load_runtime()
            c.load_shell_env()
            assert c.outer.inner.hooray == "json"

        def runtime_overrides_project(self):
            c = Config(
                runtime_path=join(CONFIGS_PATH, "json", "invoke.json"),
                project_location=join(CONFIGS_PATH, "yaml"),
            )
            c.load_runtime()
            c.load_project()
            assert c.outer.inner.hooray == "json"

        def runtime_overrides_user(self):
            c = Config(
                runtime_path=join(CONFIGS_PATH, "json", "invoke.json"),
                user_prefix=join(CONFIGS_PATH, "yaml/"),
            )
            c.load_runtime()
            assert c.outer.inner.hooray == "json"

        def runtime_overrides_systemwide(self):
            c = Config(
                runtime_path=join(CONFIGS_PATH, "json", "invoke.json"),
                system_prefix=join(CONFIGS_PATH, "yaml/"),
            )
            c.load_runtime()
            assert c.outer.inner.hooray == "json"

        def runtime_overrides_collection(self):
            c = Config(runtime_path=join(CONFIGS_PATH, "json", "invoke.json"))
            c.load_collection({"outer": {"inner": {"hooray": "defaults"}}})
            c.load_runtime()
            assert c.outer.inner.hooray == "json"

        def cli_overrides_override_all(self):
            "CLI-driven overrides win vs all other layers"
            # TODO: expand into more explicit tests like the above? meh
            c = Config(
                overrides={"outer": {"inner": {"hooray": "overrides"}}},
                runtime_path=join(CONFIGS_PATH, "json", "invoke.json"),
            )
            c.load_runtime()
            assert c.outer.inner.hooray == "overrides"

        def yaml_prevents_yml_json_or_python(self):
            c = Config(system_prefix=join(CONFIGS_PATH, "all-four/"))
            assert "json-only" not in c
            assert "python_only" not in c
            assert "yml-only" not in c
            assert "yaml-only" in c
            assert c.shared == "yaml-value"

        def yml_prevents_json_or_python(self):
            c = Config(system_prefix=join(CONFIGS_PATH, "three-of-em/"))
            assert "json-only" not in c
            assert "python_only" not in c
            assert "yml-only" in c
            assert c.shared == "yml-value"

        def json_prevents_python(self):
            c = Config(system_prefix=join(CONFIGS_PATH, "json-and-python/"))
            assert "python_only" not in c
            assert "json-only" in c
            assert c.shared == "json-value"

    class clone:
        def preserves_basic_members(self):
            c1 = Config(
                defaults={"key": "default"},
                overrides={"key": "override"},
                system_prefix="global",
                user_prefix="user",
                project_location="project",
                runtime_path="runtime.yaml",
            )
            c2 = c1.clone()
            # NOTE: expecting identical defaults also implicitly tests that
            # clone() passes in defaults= instead of doing an empty init +
            # copy. (When that is not the case, we end up with
            # global_defaults() being rerun and re-added to _defaults...)
            assert c2._defaults == c1._defaults
            assert c2._defaults is not c1._defaults
            assert c2._overrides == c1._overrides
            assert c2._overrides is not c1._overrides
            assert c2._system_prefix == c1._system_prefix
            assert c2._user_prefix == c1._user_prefix
            assert c2._project_prefix == c1._project_prefix
            assert c2.prefix == c1.prefix
            assert c2.file_prefix == c1.file_prefix
            assert c2.env_prefix == c1.env_prefix
            assert c2._runtime_path == c1._runtime_path

        def preserves_merged_config(self):
            c = Config(
                defaults={"key": "default"}, overrides={"key": "override"}
            )
            assert c.key == "override"
            assert c._defaults["key"] == "default"
            c2 = c.clone()
            assert c2.key == "override"
            assert c2._defaults["key"] == "default"
            assert c2._overrides["key"] == "override"

        def preserves_file_data(self):
            c = Config(system_prefix=join(CONFIGS_PATH, "yaml/"))
            assert c.outer.inner.hooray == "yaml"
            c2 = c.clone()
            assert c2.outer.inner.hooray == "yaml"
            assert c2._system == {"outer": {"inner": {"hooray": "yaml"}}}

        @patch.object(
            Config,
            "_load_yaml",
            return_value={"outer": {"inner": {"hooray": "yaml"}}},
        )
        def does_not_reload_file_data(self, load_yaml):
            path = join(CONFIGS_PATH, "yaml/")
            c = Config(system_prefix=path)
            c2 = c.clone()
            assert c2.outer.inner.hooray == "yaml"
            # Crummy way to say "only got called with this specific invocation
            # one time" (since assert_calls_with gets mad about other
            # invocations w/ different args)
            calls = load_yaml.call_args_list
            my_call = call("{}invoke.yaml".format(path))
            try:
                calls.remove(my_call)
                assert my_call not in calls
            except ValueError:
                err = "{} not found in {} even once!"
                assert False, err.format(my_call, calls)

        def preserves_env_data(self):
            os.environ["INVOKE_FOO"] = "bar"
            c = Config(defaults={"foo": "notbar"})
            c.load_shell_env()
            c2 = c.clone()
            assert c2.foo == "bar"

        def works_correctly_when_subclassed(self):
            # Because sometimes, implementation #1 is really naive!
            class MyConfig(Config):
                pass

            c = MyConfig()
            assert isinstance(c, MyConfig)  # sanity
            c2 = c.clone()
            assert isinstance(c2, MyConfig)  # actual test

        class into_kwarg:
            "'into' kwarg"

            def is_not_required(self):
                c = Config(defaults={"meh": "okay"})
                c2 = c.clone()
                assert c2.meh == "okay"

            def raises_TypeError_if_value_is_not_Config_subclass(self):
                try:
                    Config().clone(into=17)
                except TypeError:
                    pass
                else:
                    assert False, "Non-class obj did not raise TypeError!"

                class Foo(object):
                    pass

                try:
                    Config().clone(into=Foo)
                except TypeError:
                    pass
                else:
                    assert False, "Non-subclass did not raise TypeError!"

            def resulting_clones_are_typed_as_new_class(self):
                class MyConfig(Config):
                    pass

                c = Config()
                c2 = c.clone(into=MyConfig)
                assert type(c2) is MyConfig

            def non_conflicting_values_are_merged(self):
                # NOTE: this is really just basic clone behavior.
                class MyConfig(Config):
                    @staticmethod
                    def global_defaults():
                        orig = Config.global_defaults()
                        orig["new"] = {"data": "ohai"}
                        return orig

                c = Config(defaults={"other": {"data": "hello"}})
                c["runtime"] = {"modification": "sup"}
                c2 = c.clone(into=MyConfig)
                # New default data from MyConfig present
                assert c2.new.data == "ohai"
                # As well as old default data from the cloned instance
                assert c2.other.data == "hello"
                # And runtime user mods from the cloned instance
                assert c2.runtime.modification == "sup"

        def does_not_deepcopy(self):
            c = Config(
                defaults={
                    # Will merge_dicts happily
                    "oh": {"dear": {"god": object()}},
                    # And shallow-copy compound values
                    "shallow": {"objects": ["copy", "okay"]},
                    # Will preserve refrences to the innermost dict, sadly. Not
                    # much we can do without incurring deepcopy problems (or
                    # reimplementing it entirely)
                    "welp": {"cannot": ["have", {"everything": "we want"}]},
                }
            )
            c2 = c.clone()
            # Basic identity
            assert c is not c2, "Clone had same identity as original!"
            # Dicts get recreated
            assert c.oh is not c2.oh, "Top level key had same identity!"
            assert (
                c.oh.dear is not c2.oh.dear
            ), "Midlevel key had same identity!"  # noqa
            # Basic values get copied
            err = "Leaf object() had same identity!"
            assert c.oh.dear.god is not c2.oh.dear.god, err
            assert c.shallow.objects == c2.shallow.objects
            err = "Shallow list had same identity!"
            assert c.shallow.objects is not c2.shallow.objects, err
            # Deeply nested non-dict objects are stil problematic, oh well
            err = "Huh, a deeply nested dict-in-a-list had different identity?"
            assert c.welp.cannot[1] is c2.welp.cannot[1], err
            err = "Huh, a deeply nested dict-in-a-list value had different identity?"  # noqa
            assert (
                c.welp.cannot[1]["everything"]
                is c2.welp.cannot[1]["everything"]
            ), err  # noqa

    def can_be_pickled(self):
        c = Config(overrides={"foo": {"bar": {"biz": ["baz", "buzz"]}}})
        c2 = pickle.loads(pickle.dumps(c))
        assert c == c2
        assert c is not c2
        assert c.foo.bar.biz is not c2.foo.bar.biz


# NOTE: merge_dicts has its own very low level unit tests in its own file



from invoke import task


@task
def no_docstring(c):
    pass


@task
def task_one(c):
    """
    Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam id dictum

    risus. Nulla lorem justo, sagittis in volutpat eget
    """


@task
def task_two(c):
    """
    Nulla eget ultrices ante. Curabitur sagittis commodo posuere. Duis dapibus

    facilisis, lacus et dapibus rutrum, lectus turpis egestas dui
    """



"Tasks for compiling static code and assets."

from invoke import task, Collection

from . import docs, python


@task(name="all", aliases=["everything"], default=True)
def all_(c):
    "Build all necessary artifacts."
    pass


@task(aliases=["ext"])
def c_ext(c):
    "Build our internal C extension."
    pass


@task
def zap(c):
    "A silly way to clean."
    pass


ns = Collection(all_, c_ext, zap, docs, python)



from pytest import raises

from invoke.config import merge_dicts, copy_dict, AmbiguousMergeError


class merge_dicts_:
    # NOTE: don't usually like doing true unit tests of low level plumbing -
    # prefer to infer it's all working by examining higher level behavior - but
    # sometimes it's necessary to more easily stamp out certain bugs.

    def merging_data_onto_empty_dict(self):
        d1 = {}
        d2 = {"foo": "bar"}
        merge_dicts(d1, d2)
        assert d1 == d2

    def updating_with_None_acts_like_merging_empty_dict(self):
        # When bug present, AttributeError is raised on a None.items()
        d1 = {"my": "data"}
        d2 = None
        merge_dicts(d1, d2)
        assert d1 == {"my": "data"}

    def orthogonal_data_merges(self):
        d1 = {"foo": "bar"}
        d2 = {"biz": "baz"}
        merge_dicts(d1, d2)
        assert d1 == {"foo": "bar", "biz": "baz"}

    def updates_arg_values_win(self):
        d1 = {"foo": "bar"}
        d2 = {"foo": "notbar"}
        merge_dicts(d1, d2)
        assert d1 == {"foo": "notbar"}

    def non_dict_type_mismatch_overwrites_ok(self):
        d1 = {"foo": "bar"}
        d2 = {"foo": [1, 2, 3]}
        merge_dicts(d1, d2)
        assert d1 == {"foo": [1, 2, 3]}

    def merging_dict_into_nondict_raises_error(self):
        # TODO: or...should it?! If a user really wants to take a pre-existing
        # config path and make it 'deeper' by overwriting e.g. a string with a
        # dict of strings (or whatever)...should they be allowed to?
        d1 = {"foo": "bar"}
        d2 = {"foo": {"uh": "oh"}}
        with raises(AmbiguousMergeError):
            merge_dicts(d1, d2)

    def merging_nondict_into_dict_raises_error(self):
        d1 = {"foo": {"uh": "oh"}}
        d2 = {"foo": "bar"}
        with raises(AmbiguousMergeError):
            merge_dicts(d1, d2)

    def nested_leaf_values_merge_ok(self):
        d1 = {"foo": {"bar": {"biz": "baz"}}}
        d2 = {"foo": {"bar": {"biz": "notbaz"}}}
        merge_dicts(d1, d2)
        assert d1 == {"foo": {"bar": {"biz": "notbaz"}}}

    def mixed_branch_levels_merges_ok(self):
        d1 = {"foo": {"bar": {"biz": "baz"}}, "meh": 17, "myown": "ok"}
        d2 = {"foo": {"bar": {"biz": "notbaz"}}, "meh": 25}
        merge_dicts(d1, d2)
        expected = {
            "foo": {"bar": {"biz": "notbaz"}},
            "meh": 25,
            "myown": "ok",
        }
        assert d1 == expected

    def dict_value_merges_are_not_references(self):
        core = {}
        coll = {"foo": {"bar": {"biz": "coll value"}}}
        proj = {"foo": {"bar": {"biz": "proj value"}}}
        # Initial merge - when bug present, this sets core['foo'] to the entire
        # 'foo' dict in 'proj' as a reference - meaning it 'links' back to the
        # 'proj' dict whenever other things are merged into it
        merge_dicts(core, proj)
        assert core == {"foo": {"bar": {"biz": "proj value"}}}
        assert proj["foo"]["bar"]["biz"] == "proj value"
        # Identity tests can also prove the bug early
        assert (
            core["foo"] is not proj["foo"]
        ), "Core foo is literally proj foo!"  # noqa
        # Subsequent merge - just overwrites leaf values this time (thus no
        # real change, but this is what real config merge code does, so why
        # not)
        merge_dicts(core, proj)
        assert core == {"foo": {"bar": {"biz": "proj value"}}}
        assert proj["foo"]["bar"]["biz"] == "proj value"
        # The problem merge - when bug present, core['foo'] references 'foo'
        # inside 'proj', so this ends up tweaking "core" but it actually
        # affects "proj" as well!
        merge_dicts(core, coll)
        # Expect that the core dict got the update from 'coll'...
        assert core == {"foo": {"bar": {"biz": "coll value"}}}
        # BUT that 'proj' remains UNTOUCHED
        assert proj["foo"]["bar"]["biz"] == "proj value"

    def merge_file_types_by_reference(self):
        with open(__file__) as fd:
            d1 = {}
            d2 = {"foo": fd}
            merge_dicts(d1, d2)
            assert d1["foo"].closed is False


class copy_dict_:
    def returns_deep_copy_of_given_dict(self):
        # NOTE: not actual deepcopy...
        source = {"foo": {"bar": {"biz": "baz"}}}
        copy = copy_dict(source)
        assert copy["foo"]["bar"] == source["foo"]["bar"]
        assert copy["foo"]["bar"] is not source["foo"]["bar"]
        copy["foo"]["bar"]["biz"] = "notbaz"
        assert source["foo"]["bar"]["biz"] == "baz"



from pytest import raises

from invoke.parser import Parser, Context, Argument, ParseError


class Parser_:
    def can_take_initial_context(self):
        c = Context()
        p = Parser(initial=c)
        assert p.initial == c

    def can_take_initial_and_other_contexts(self):
        c1 = Context("foo")
        c2 = Context("bar")
        p = Parser(initial=Context(), contexts=[c1, c2])
        assert p.contexts["foo"] == c1
        assert p.contexts["bar"] == c2

    def can_take_just_other_contexts(self):
        c = Context("foo")
        p = Parser(contexts=[c])
        assert p.contexts["foo"] == c

    def can_take_just_contexts_as_non_keyword_arg(self):
        c = Context("foo")
        p = Parser([c])
        assert p.contexts["foo"] == c

    def raises_ValueError_for_unnamed_Contexts_in_contexts(self):
        with raises(ValueError):
            Parser(initial=Context(), contexts=[Context()])

    def raises_error_for_context_name_clashes(self):
        with raises(ValueError):
            Parser(contexts=(Context("foo"), Context("foo")))

    def raises_error_for_context_alias_and_name_clashes(self):
        with raises(ValueError):
            Parser((Context("foo", aliases=("bar",)), Context("bar")))

    def raises_error_for_context_name_and_alias_clashes(self):
        # I.e. inverse of the above, which is a different code path.
        with raises(ValueError):
            Parser((Context("foo"), Context("bar", aliases=("foo",))))

    def takes_ignore_unknown_kwarg(self):
        Parser(ignore_unknown=True)

    def ignore_unknown_defaults_to_False(self):
        assert Parser().ignore_unknown is False

    class parse_argv:
        def parses_sys_argv_style_list_of_strings(self):
            "parses sys.argv-style list of strings"
            # Doesn't-blow-up tests FTL
            mytask = Context(name="mytask")
            mytask.add_arg("arg")
            p = Parser(contexts=[mytask])
            p.parse_argv(["mytask", "--arg", "value"])

        def returns_only_contexts_mentioned(self):
            task1 = Context("mytask")
            task2 = Context("othertask")
            result = Parser((task1, task2)).parse_argv(["othertask"])
            assert len(result) == 1
            assert result[0].name == "othertask"

        def raises_error_if_unknown_contexts_found(self):
            with raises(ParseError):
                Parser().parse_argv(["foo", "bar"])

        def unparsed_does_not_share_state(self):
            r = Parser(ignore_unknown=True).parse_argv(["self"])
            assert r.unparsed == ["self"]
            r2 = Parser(ignore_unknown=True).parse_argv(["contained"])
            assert r.unparsed == ["self"]  # NOT ['self', 'contained']
            assert r2.unparsed == ["contained"]  # NOT ['self', 'contained']

        def ignore_unknown_returns_unparsed_argv_instead(self):
            r = Parser(ignore_unknown=True).parse_argv(["foo", "bar", "--baz"])
            assert r.unparsed == ["foo", "bar", "--baz"]

        def ignore_unknown_does_not_mutate_rest_of_argv(self):
            p = Parser([Context("ugh")], ignore_unknown=True)
            r = p.parse_argv(["ugh", "what", "-nowai"])
            # NOT: ['what', '-n', '-w', '-a', '-i']
            assert r.unparsed == ["what", "-nowai"]

        def always_includes_initial_context_if_one_was_given(self):
            # Even if no core/initial flags were seen
            t1 = Context("t1")
            init = Context()
            result = Parser((t1,), initial=init).parse_argv(["t1"])
            assert result[0].name is None
            assert result[1].name == "t1"

        def returned_contexts_are_in_order_given(self):
            t1, t2 = Context("t1"), Context("t2")
            r = Parser((t1, t2)).parse_argv(["t2", "t1"])
            assert [x.name for x in r] == ["t2", "t1"]

        def returned_context_member_arguments_contain_given_values(self):
            c = Context("mytask", args=(Argument("boolean", kind=bool),))
            result = Parser((c,)).parse_argv(["mytask", "--boolean"])
            assert result[0].args["boolean"].value is True

        def inverse_bools_get_set_correctly(self):
            arg = Argument("myarg", kind=bool, default=True)
            c = Context("mytask", args=(arg,))
            r = Parser((c,)).parse_argv(["mytask", "--no-myarg"])
            assert r[0].args["myarg"].value is False

        def arguments_which_take_values_get_defaults_overridden_correctly(
            self,
        ):  # noqa
            args = (Argument("arg", kind=str), Argument("arg2", kind=int))
            c = Context("mytask", args=args)
            argv = ["mytask", "--arg", "myval", "--arg2", "25"]
            result = Parser((c,)).parse_argv(argv)
            assert result[0].args["arg"].value == "myval"
            assert result[0].args["arg2"].value == 25

        def returned_arguments_not_given_contain_default_values(self):
            # I.e. a Context with args A and B, invoked with no mention of B,
            # should result in B existing in the result, with its default value
            # intact, and not e.g. None, or the arg not existing.
            a = Argument("name", kind=str)
            b = Argument("age", default=7)
            c = Context("mytask", args=(a, b))
            Parser((c,)).parse_argv(["mytask", "--name", "blah"])
            assert c.args["age"].value == 7

        def returns_remainder(self):
            "returns -- style remainder string chunk"
            r = Parser((Context("foo"),)).parse_argv(
                ["foo", "--", "bar", "biz"]
            )
            assert r.remainder == "bar biz"

        def clones_initial_context(self):
            a = Argument("foo", kind=bool)
            assert a.value is None
            c = Context(args=(a,))
            p = Parser(initial=c)
            assert p.initial is c
            r = p.parse_argv(["--foo"])
            assert p.initial is c
            c2 = r[0]
            assert c2 is not c
            a2 = c2.args["foo"]
            assert a2 is not a
            assert a.value is None
            assert a2.value is True

        def clones_noninitial_contexts(self):
            a = Argument("foo")
            assert a.value is None
            c = Context(name="mytask", args=(a,))
            p = Parser(contexts=(c,))
            assert p.contexts["mytask"] is c
            r = p.parse_argv(["mytask", "--foo", "val"])
            assert p.contexts["mytask"] is c
            c2 = r[0]
            assert c2 is not c
            a2 = c2.args["foo"]
            assert a2 is not a
            assert a.value is None
            assert a2.value == "val"

        class parsing_errors:
            def setup_method(self):
                self.p = Parser([Context(name="foo", args=[Argument("bar")])])

            def missing_flag_values_raise_ParseError(self):
                with raises(ParseError):
                    self.p.parse_argv(["foo", "--bar"])

            def attaches_context_to_ParseErrors(self):
                try:
                    self.p.parse_argv(["foo", "--bar"])
                except ParseError as e:
                    assert e.context is not None

            def attached_context_is_None_outside_contexts(self):
                try:
                    Parser().parse_argv(["wat"])
                except ParseError as e:
                    assert e.context is None

        class positional_arguments:
            def _basic(self):
                arg = Argument("pos", positional=True)
                mytask = Context(name="mytask", args=[arg])
                return Parser(contexts=[mytask])

            def single_positional_arg(self):
                r = self._basic().parse_argv(["mytask", "posval"])
                assert r[0].args["pos"].value == "posval"

            def omitted_positional_arg_raises_ParseError(self):
                try:
                    self._basic().parse_argv(["mytask"])
                except ParseError as e:
                    expected = "'mytask' did not receive required positional arguments: 'pos'"  # noqa
                    assert str(e) == expected
                else:
                    assert False, "Did not raise ParseError!"

            def omitted_positional_args_raises_ParseError(self):
                try:
                    arg = Argument("pos", positional=True)
                    arg2 = Argument("morepos", positional=True)
                    mytask = Context(name="mytask", args=[arg, arg2])
                    Parser(contexts=[mytask]).parse_argv(["mytask"])
                except ParseError as e:
                    expected = "'mytask' did not receive required positional arguments: 'pos', 'morepos'"  # noqa
                    assert str(e) == expected
                else:
                    assert False, "Did not raise ParseError!"

            def positional_args_eat_otherwise_valid_context_names(self):
                mytask = Context(
                    "mytask",
                    args=[
                        Argument("pos", positional=True),
                        Argument("nonpos", default="default"),
                    ],
                )
                Context("lolwut")
                result = Parser([mytask]).parse_argv(["mytask", "lolwut"])
                r = result[0]
                assert r.args["pos"].value == "lolwut"
                assert r.args["nonpos"].value == "default"
                assert len(result) == 1  # Not 2

            def positional_args_can_still_be_given_as_flags(self):
                # AKA "positional args can come anywhere in the context"
                pos1 = Argument("pos1", positional=True)
                pos2 = Argument("pos2", positional=True)
                nonpos = Argument("nonpos", positional=False, default="lol")
                mytask = Context("mytask", args=[pos1, pos2, nonpos])
                assert mytask.positional_args == [pos1, pos2]
                r = Parser([mytask]).parse_argv(
                    [
                        "mytask",
                        "--nonpos",
                        "wut",
                        "--pos2",
                        "pos2val",
                        "pos1val",
                    ]
                )[0]
                assert r.args["pos1"].value == "pos1val"
                assert r.args["pos2"].value == "pos2val"
                assert r.args["nonpos"].value == "wut"

        class equals_signs:
            def _compare(self, argname, invoke, value):
                c = Context("mytask", args=(Argument(argname, kind=str),))
                r = Parser((c,)).parse_argv(["mytask", invoke])
                assert r[0].args[argname].value == value

            def handles_equals_style_long_flags(self):
                self._compare("foo", "--foo=bar", "bar")

            def handles_equals_style_short_flags(self):
                self._compare("f", "-f=bar", "bar")

            def does_not_require_escaping_equals_signs_in_value(self):
                self._compare("f", "-f=biz=baz", "biz=baz")

        def handles_multiple_boolean_flags_per_context(self):
            c = Context(
                "mytask",
                args=(Argument("foo", kind=bool), Argument("bar", kind=bool)),
            )
            r = Parser([c]).parse_argv(["mytask", "--foo", "--bar"])
            a = r[0].args
            assert a.foo.value is True
            assert a.bar.value is True

    class optional_arg_values:
        def setup_method(self):
            self.parser = self._parser()

        def _parser(self, arguments=None):
            if arguments is None:
                arguments = (
                    Argument(
                        names=("foo", "f"), optional=True, default="mydefault"
                    ),
                )
            self.context = Context("mytask", args=arguments)
            self.parser = Parser([self.context])
            return self.parser

        def _parse(self, argstr, parser=None):
            parser = parser or self.parser
            return parser.parse_argv(["mytask"] + argstr.split())

        def _expect(self, argstr, expected, parser=None):
            result = self._parse(argstr, parser)
            assert result[0].args.foo.value == expected

        def no_value_becomes_True_not_default_value(self):
            self._expect("--foo", True)
            self._expect("-f", True)

        def value_given_gets_preserved_normally(self):
            for argstr in (
                "--foo whatever",
                "--foo=whatever",
                "-f whatever",
                "-f=whatever",
            ):
                self._expect(argstr, "whatever")

        def not_given_at_all_uses_default_value(self):
            self._expect("", "mydefault")

        class ambiguity_sanity_checks:
            def _test_for_ambiguity(self, invoke, parser=None):
                msg = "is ambiguous"
                try:
                    self._parse(invoke, parser or self.parser)
                # Expected result
                except ParseError as e:
                    assert msg in str(e)
                # No exception occurred at all? Bollocks.
                else:
                    assert False
                # Any other exceptions will naturally cause failure here.

            def unfilled_posargs(self):
                p = self._parser(
                    (
                        Argument("foo", optional=True),
                        Argument("bar", positional=True),
                    )
                )
                self._test_for_ambiguity("--foo uhoh", p)

            def no_ambiguity_if_option_val_already_given(self):
                p = self._parser(
                    (
                        Argument("foo", optional=True),
                        Argument("bar", kind=bool),
                    )
                )
                # This should NOT raise a ParseError.
                result = self._parse("--foo hello --bar", p)
                assert result[0].args["foo"].value == "hello"
                assert result[0].args["bar"].value is True

            def valid_argument_is_NOT_ambiguous(self):
                # The one exception that proves the rule?
                self._parser((Argument("foo", optional=True), Argument("bar")))
                for form in ("--bar barval", "--bar=barval"):
                    result = self._parse("--foo {}".format(form))
                    assert len(result) == 1
                    args = result[0].args
                    assert args["foo"].value is True
                    assert args["bar"].value == "barval"

            def valid_flaglike_argument_is_NOT_ambiguous(self):
                # The OTHER exception that proves the rule?
                self._parser(
                    (
                        Argument("foo", optional=True),
                        Argument("bar", kind=bool),
                    )
                )
                result = self._parse("--foo --bar")
                assert len(result) == 1
                args = result[0].args
                assert args["foo"].value is True
                assert args["bar"].value is True

            def invalid_flaglike_value_is_stored_as_value(self):
                self._parser((Argument("foo", optional=True),))
                result = self._parse("--foo --bar")
                assert result[0].args["foo"].value == "--bar"

            def task_name(self):
                # mytask --foo myothertask
                c1 = Context("mytask", args=(Argument("foo", optional=True),))
                c2 = Context("othertask")
                p = Parser([c1, c2])
                self._test_for_ambiguity("--foo othertask", p)

    class list_type_arguments:
        "list-type (iterable) arguments"

        def _parse(self, *args):
            c = Context("mytask", args=(Argument("mylist", kind=list),))
            argv = ["mytask"] + list(args)
            return Parser([c]).parse_argv(argv)[0].args.mylist.value

        def can_be_given_no_times_resulting_in_default_empty_list(self):
            assert self._parse() == []

        def given_once_becomes_single_item_list(self):
            assert self._parse("--mylist", "foo") == ["foo"]

        def given_N_times_becomes_list_of_len_N(self):
            expected = ["foo", "bar", "biz"]
            got = self._parse(
                "--mylist", "foo", "--mylist", "bar", "--mylist", "biz"
            )
            assert got == expected

        def iterables_work_correctly_outside_a_vacuum(self):
            # Undetected bug where I was primarily focused on the -vvv use
            # case...'normal' incrementables never left 'waiting for value'
            # state in the parser! so _subsequent_ task names & such never got
            # parsed right, always got appended to the list.
            c = Context("mytask", args=[Argument("mylist", kind=list)])
            c2 = Context("othertask")
            argv = [
                "mytask",
                "--mylist",
                "val",
                "--mylist",
                "val2",
                "othertask",
            ]
            result = Parser([c, c2]).parse_argv(argv)
            # When bug present, result only has one context (for 'mytask') and
            # its 'mylist' consists of ['val', 'val2', 'othertask']. (the
            # middle '--mylist' was handled semi-correctly.)
            mylist = result[0].args.mylist.value
            assert mylist == ["val", "val2"]
            contexts = len(result)
            err = "Got {} parse context results instead of 2!".format(contexts)
            assert contexts == 2, err
            assert result[1].name == "othertask"

    class task_repetition:
        def is_happy_to_handle_same_task_multiple_times(self):
            task1 = Context("mytask")
            result = Parser((task1,)).parse_argv(["mytask", "mytask"])
            assert len(result) == 2
            for x in result:
                assert x.name == "mytask"

        def task_args_work_correctly(self):
            task1 = Context("mytask", args=(Argument("meh"),))
            result = Parser((task1,)).parse_argv(
                ["mytask", "--meh", "mehval1", "mytask", "--meh", "mehval2"]
            )
            assert result[0].args.meh.value == "mehval1"
            assert result[1].args.meh.value == "mehval2"

    class per_task_core_flags:
        class general:
            def _echo(self):
                return Argument("echo", kind=bool, default=False)

            def core_flags_work_normally_when_no_conflict(self):
                # Initial parse context with an --echo, plus a no-args task
                initial = Context(args=[self._echo()])
                task1 = Context("mytask")
                parser = Parser(initial=initial, contexts=[task1])
                # Call with --echo in the per-task context, expect the core
                # context got updated (vs an error)
                result = parser.parse_argv(["mytask", "--echo"])
                assert result[0].args.echo.value is True

            def when_conflict_per_task_args_win_out(self):
                # Initial parse context with an --echo, plus task w/ same
                initial = Context(args=[self._echo()])
                task1 = Context("mytask", args=[self._echo()])
                parser = Parser(initial=initial, contexts=[task1])
                # Call with --echo in the per-task context, expect the task
                # context got updated, and not core.
                result = parser.parse_argv(["mytask", "--echo"])
                assert result[0].args.echo.value is False
                assert result[1].args.echo.value is True

            def value_requiring_core_flags_also_work_correctly(self):
                "value-requiring core flags also work correctly"
                initial = Context(args=[Argument("hide")])
                task1 = Context("mytask")
                parser = Parser(initial=initial, contexts=[task1])
                result = parser.parse_argv(["mytask", "--hide", "both"])
                assert result[0].args.hide.value == "both"

        class edge_cases:
            def core_bool_but_per_task_string(self):
                # Initial parse context with bool --hide, and a task with a
                # regular (string) --hide
                initial = Context(
                    args=[Argument("hide", kind=bool, default=False)]
                )
                task1 = Context("mytask", args=[Argument("hide")])
                parser = Parser(initial=initial, contexts=[task1])
                # Expect that, because the task's version wins, we're able to
                # call it with a value. (If there were weird bugs where the
                # core flag informed the parsing, this would fail.)
                result = parser.parse_argv(["mytask", "--hide", "both"])
                assert result[0].args.hide.value is False
                assert result[1].args.hide.value == "both"

        class help_treats_context_name_as_its_value:
            def by_itself_base_case(self):
                task1 = Context("mytask")
                init = Context(args=[Argument("help", optional=True)])
                parser = Parser(initial=init, contexts=[task1])
                result = parser.parse_argv(["mytask", "--help"])
                assert len(result) == 2
                assert result[0].args.help.value == "mytask"
                assert "help" not in result[1].args

            def other_tokens_afterwards_raise_parse_errors(self):
                # NOTE: this is because of the special-casing where we supply
                # the task name as the value when the flag is literally named
                # "help".
                task1 = Context("mytask")
                init = Context(args=[Argument("help", optional=True)])
                parser = Parser(initial=init, contexts=[task1])
                with raises(ParseError, match=r".*foobar.*"):
                    parser.parse_argv(["mytask", "--help", "foobar"])


class ParseResult_:
    "ParseResult"

    def setup_method(self):
        self.context = Context(
            "mytask", args=(Argument("foo", kind=str), Argument("bar"))
        )
        argv = ["mytask", "--foo", "foo-val", "--", "my", "remainder"]
        self.result = Parser((self.context,)).parse_argv(argv)

    def acts_as_a_list_of_parsed_contexts(self):
        assert len(self.result) == 1
        assert self.result[0].name == "mytask"

    def exhibits_remainder_attribute(self):
        assert self.result.remainder == "my remainder"



import errno
import os
import signal
import struct
import sys
import termios
import threading
import types
from io import StringIO
from io import BytesIO
from itertools import chain, repeat

from pytest import raises, skip
from pytest_relaxed import trap
from unittest.mock import patch, Mock, call

from invoke import (
    CommandTimedOut,
    Config,
    Context,
    Failure,
    Local,
    Promise,
    Responder,
    Result,
    Runner,
    StreamWatcher,
    SubprocessPipeError,
    ThreadException,
    UnexpectedExit,
    WatcherError,
)
from invoke.runners import default_encoding
from invoke.terminals import WINDOWS

from _util import (
    mock_subprocess,
    mock_pty,
    skip_if_windows,
    _Dummy,
    _KeyboardInterruptingRunner,
    OhNoz,
    _,
)


class _RaisingWatcher(StreamWatcher):
    def submit(self, stream):
        raise WatcherError("meh")


class _GenericException(Exception):
    pass


class _GenericExceptingRunner(_Dummy):
    def wait(self):
        raise _GenericException


def _run(*args, **kwargs):
    klass = kwargs.pop("klass", _Dummy)
    settings = kwargs.pop("settings", {})
    context = Context(config=Config(overrides=settings))
    return klass(context).run(*args, **kwargs)


def _runner(out="", err="", **kwargs):
    klass = kwargs.pop("klass", _Dummy)
    runner = klass(Context(config=Config(overrides=kwargs)))
    if "exits" in kwargs:
        runner.returncode = Mock(return_value=kwargs.pop("exits"))
    out_file = BytesIO(out.encode())
    err_file = BytesIO(err.encode())
    runner.read_proc_stdout = out_file.read
    runner.read_proc_stderr = err_file.read
    return runner


def _expect_platform_shell(shell):
    if WINDOWS:
        assert shell.endswith("cmd.exe")
    else:
        assert shell == "/bin/bash"


def _make_tcattrs(cc_is_ints=True, echo=False):
    # Set up the control character sub-array; it's technically platform
    # dependent so we need to be dynamic.
    # NOTE: setting this up so we can test both potential values for
    # the 'cc' members...docs say ints, reality says one-byte
    # bytestrings...
    cc_base = [None] * (max(termios.VMIN, termios.VTIME) + 1)
    cc_ints, cc_bytes = cc_base.copy(), cc_base.copy()
    cc_ints[termios.VMIN], cc_ints[termios.VTIME] = 1, 0
    cc_bytes[termios.VMIN], cc_bytes[termios.VTIME] = b"\x01", b"\x00"
    # Set tcgetattr to look like it's already cbroken...
    attrs = [
        # iflag, oflag, cflag - don't care
        None,
        None,
        None,
        # lflag needs to have ECHO and ICANON unset
        ~(termios.ECHO | termios.ICANON),
        # ispeed, ospeed - don't care
        None,
        None,
        # cc - care about its VMIN and VTIME members.
        cc_ints if cc_is_ints else cc_bytes,
    ]
    # Undo the ECHO unset if caller wants this to look like a non-cbroken term
    if echo:
        attrs[3] = attrs[3] | termios.ECHO
    return attrs


class _TimingOutRunner(_Dummy):
    @property
    def timed_out(self):
        return True


class Runner_:
    _stop_methods = ["generate_result", "stop"]

    # NOTE: these copies of _run and _runner form the base case of "test Runner
    # subclasses via self._run/_runner helpers" functionality. See how e.g.
    # Local_ uses the same approach but bakes in the dummy class used.
    def _run(self, *args, **kwargs):
        return _run(*args, **kwargs)

    def _runner(self, *args, **kwargs):
        return _runner(*args, **kwargs)

    def _mock_stdin_writer(self):
        """
        Return new _Dummy subclass whose write_proc_stdin() method is a mock.
        """

        class MockedStdin(_Dummy):
            pass

        MockedStdin.write_proc_stdin = Mock()
        return MockedStdin

    class init:
        "__init__"

        def takes_a_context_instance(self):
            c = Context()
            assert Runner(c).context == c

        def context_instance_is_required(self):
            with raises(TypeError):
                Runner()

    class run:
        def handles_invalid_kwargs_like_any_other_function(self):
            try:
                self._run(_, nope_noway_nohow="as if")
            except TypeError as e:
                assert "got an unexpected keyword argument" in str(e)
            else:
                assert False, "Invalid run() kwarg didn't raise TypeError"

    class warn:
        def honors_config(self):
            runner = self._runner(run={"warn": True}, exits=1)
            # Doesn't raise Failure -> all good
            runner.run(_)

        def kwarg_beats_config(self):
            runner = self._runner(run={"warn": False}, exits=1)
            # Doesn't raise Failure -> all good
            runner.run(_, warn=True)

        def does_not_apply_to_watcher_errors(self):
            runner = self._runner(out="stuff")
            try:
                watcher = _RaisingWatcher()
                runner.run(_, watchers=[watcher], warn=True, hide=True)
            except Failure as e:
                assert isinstance(e.reason, WatcherError)
            else:
                assert False, "Did not raise Failure for WatcherError!"

        def does_not_apply_to_timeout_errors(self):
            with raises(CommandTimedOut):
                self._runner(klass=_TimingOutRunner).run(
                    _, timeout=1, warn=True
                )

    class hide:
        @trap
        def honors_config(self):
            runner = self._runner(out="stuff", run={"hide": True})
            r = runner.run(_)
            assert r.stdout == "stuff"
            assert sys.stdout.getvalue() == ""

        @trap
        def kwarg_beats_config(self):
            runner = self._runner(out="stuff")
            r = runner.run(_, hide=True)
            assert r.stdout == "stuff"
            assert sys.stdout.getvalue() == ""

    class pty:
        def pty_defaults_to_off(self):
            assert self._run(_).pty is False

        def honors_config(self):
            runner = self._runner(run={"pty": True})
            assert runner.run(_).pty is True

        def kwarg_beats_config(self):
            runner = self._runner(run={"pty": False})
            assert runner.run(_, pty=True).pty is True

    class shell:
        def defaults_to_bash_or_cmdexe_when_pty_True(self):
            _expect_platform_shell(self._run(_, pty=True).shell)

        def defaults_to_bash_or_cmdexe_when_pty_False(self):
            _expect_platform_shell(self._run(_, pty=False).shell)

        def may_be_overridden(self):
            assert self._run(_, shell="/bin/zsh").shell == "/bin/zsh"

        def may_be_configured(self):
            runner = self._runner(run={"shell": "/bin/tcsh"})
            assert runner.run(_).shell == "/bin/tcsh"

        def kwarg_beats_config(self):
            runner = self._runner(run={"shell": "/bin/tcsh"})
            assert runner.run(_, shell="/bin/zsh").shell == "/bin/zsh"

    class env:
        def defaults_to_os_environ(self):
            assert self._run(_).env == os.environ

        def updates_when_dict_given(self):
            expected = dict(os.environ, FOO="BAR")
            assert self._run(_, env={"FOO": "BAR"}).env == expected

        def replaces_when_replace_env_True(self):
            env = self._run(_, env={"JUST": "ME"}, replace_env=True).env
            assert env == {"JUST": "ME"}

        def config_can_be_used(self):
            env = self._run(_, settings={"run": {"env": {"FOO": "BAR"}}}).env
            assert env == dict(os.environ, FOO="BAR")

        def kwarg_wins_over_config(self):
            settings = {"run": {"env": {"FOO": "BAR"}}}
            kwarg = {"FOO": "NOTBAR"}
            foo = self._run(_, settings=settings, env=kwarg).env["FOO"]
            assert foo == "NOTBAR"

    class return_value:
        def return_code(self):
            """
            Result has .return_code (and .exited) containing exit code int
            """
            runner = self._runner(exits=17)
            r = runner.run(_, warn=True)
            assert r.return_code == 17
            assert r.exited == 17

        def ok_attr_indicates_success(self):
            runner = self._runner()
            assert runner.run(_).ok is True  # default dummy retval is 0

        def ok_attr_indicates_failure(self):
            runner = self._runner(exits=1)
            assert runner.run(_, warn=True).ok is False

        def failed_attr_indicates_success(self):
            runner = self._runner()
            assert runner.run(_).failed is False  # default dummy retval is 0

        def failed_attr_indicates_failure(self):
            runner = self._runner(exits=1)
            assert runner.run(_, warn=True).failed is True

        @trap
        def stdout_attribute_contains_stdout(self):
            runner = self._runner(out="foo")
            assert runner.run(_).stdout == "foo"
            assert sys.stdout.getvalue() == "foo"

        @trap
        def stderr_attribute_contains_stderr(self):
            runner = self._runner(err="foo")
            assert runner.run(_).stderr == "foo"
            assert sys.stderr.getvalue() == "foo"

        def whether_pty_was_used(self):
            assert self._run(_).pty is False
            assert self._run(_, pty=True).pty is True

        def command_executed(self):
            assert self._run(_).command == _

        def shell_used(self):
            _expect_platform_shell(self._run(_).shell)

        def hide_param_exposed_and_normalized(self):
            assert self._run(_, hide=True).hide, "stdout" == "stderr"
            assert self._run(_, hide=False).hide == tuple()
            assert self._run(_, hide="stderr").hide == ("stderr",)

    class command_echoing:
        @trap
        def off_by_default(self):
            self._run("my command")
            assert sys.stdout.getvalue() == ""

        @trap
        def enabled_via_kwarg(self):
            self._run("my command", echo=True)
            assert "my command" in sys.stdout.getvalue()

        @trap
        def enabled_via_config(self):
            self._run("yup", settings={"run": {"echo": True}})
            assert "yup" in sys.stdout.getvalue()

        @trap
        def kwarg_beats_config(self):
            self._run("yup", echo=True, settings={"run": {"echo": False}})
            assert "yup" in sys.stdout.getvalue()

        @trap
        def uses_ansi_bold(self):
            self._run("my command", echo=True)
            # TODO: vendor & use a color module
            assert sys.stdout.getvalue() == "\x1b[1;37mmy command\x1b[0m\n"

        @trap
        def uses_custom_format(self):
            self._run(
                "my command",
                echo=True,
                settings={"run": {"echo_format": "AA{command}ZZ"}},
            )
            assert sys.stdout.getvalue() == "AAmy commandZZ\n"

    class dry_running:
        @trap
        def sets_echo_to_True(self):
            self._run("what up", settings={"run": {"dry": True}})
            assert "what up" in sys.stdout.getvalue()

        @trap
        def short_circuits_with_dummy_result(self):
            runner = self._runner(run={"dry": True})
            # Using the call to self.start() in _run_body() as a sentinel for
            # all the work beyond it.
            runner.start = Mock()
            result = runner.run(_)
            assert not runner.start.called
            assert isinstance(result, Result)
            assert result.command == _
            assert result.stdout == ""
            assert result.stderr == ""
            assert result.exited == 0
            assert result.pty is False

    class encoding:
        # NOTE: these tests just check what Runner.encoding ends up as; it's
        # difficult/impossible to mock string objects themselves to see what
        # .decode() is being given :(
        #
        # TODO: consider using truly "nonstandard"-encoded byte sequences as
        # fixtures, encoded with something that isn't compatible with UTF-8
        # (UTF-7 kinda is, so...) so we can assert that the decoded string is
        # equal to its Unicode equivalent.
        #
        # Use UTF-7 as a valid encoding unlikely to be a real default derived
        # from test-runner's locale.getpreferredencoding()
        def defaults_to_encoding_method_result(self):
            # Setup
            runner = self._runner()
            encoding = "UTF-7"
            runner.default_encoding = Mock(return_value=encoding)
            # Execution & assertion
            runner.run(_)
            runner.default_encoding.assert_called_with()
            assert runner.encoding == "UTF-7"

        def honors_config(self):
            c = Context(Config(overrides={"run": {"encoding": "UTF-7"}}))
            runner = _Dummy(c)
            runner.default_encoding = Mock(return_value="UTF-not-7")
            runner.run(_)
            assert runner.encoding == "UTF-7"

        def honors_kwarg(self):
            skip()

        def uses_locale_module_for_default_encoding(self):
            # Actually testing this highly OS/env specific stuff is very
            # error-prone; so we degrade to just testing expected function
            # calls for now :(
            with patch("invoke.runners.locale") as fake_locale:
                fake_locale.getdefaultlocale.return_value = ("meh", "UHF-8")
                fake_locale.getpreferredencoding.return_value = "FALLBACK"
                assert self._runner().default_encoding() == "FALLBACK"

        def falls_back_to_defaultlocale_when_preferredencoding_is_None(self):
            with patch("invoke.runners.locale") as fake_locale:
                fake_locale.getdefaultlocale.return_value = (None, None)
                fake_locale.getpreferredencoding.return_value = "FALLBACK"
                assert self._runner().default_encoding() == "FALLBACK"

    class output_hiding:
        @trap
        def _expect_hidden(self, hide, expect_out="", expect_err=""):
            self._runner(out="foo", err="bar").run(_, hide=hide)
            assert sys.stdout.getvalue() == expect_out
            assert sys.stderr.getvalue() == expect_err

        def both_hides_everything(self):
            self._expect_hidden("both")

        def True_hides_everything(self):
            self._expect_hidden(True)

        def out_only_hides_stdout(self):
            self._expect_hidden("out", expect_out="", expect_err="bar")

        def err_only_hides_stderr(self):
            self._expect_hidden("err", expect_out="foo", expect_err="")

        def accepts_stdout_alias_for_out(self):
            self._expect_hidden("stdout", expect_out="", expect_err="bar")

        def accepts_stderr_alias_for_err(self):
            self._expect_hidden("stderr", expect_out="foo", expect_err="")

        def None_hides_nothing(self):
            self._expect_hidden(None, expect_out="foo", expect_err="bar")

        def False_hides_nothing(self):
            self._expect_hidden(False, expect_out="foo", expect_err="bar")

        def unknown_vals_raises_ValueError(self):
            with raises(ValueError):
                self._run(_, hide="wat?")

        def unknown_vals_mention_value_given_in_error(self):
            value = "penguinmints"
            try:
                self._run(_, hide=value)
            except ValueError as e:
                msg = "Error from run(hide=xxx) did not tell user what the bad value was!"  # noqa
                msg += "\nException msg: {}".format(e)
                assert value in str(e), msg
            else:
                assert (
                    False
                ), "run() did not raise ValueError for bad hide= value"  # noqa

        def does_not_affect_capturing(self):
            assert self._runner(out="foo").run(_, hide=True).stdout == "foo"

        @trap
        def overrides_echoing(self):
            self._runner().run("invisible", hide=True, echo=True)
            assert "invisible" not in sys.stdout.getvalue()

    class output_stream_overrides:
        @trap
        def out_defaults_to_sys_stdout(self):
            "out_stream defaults to sys.stdout"
            self._runner(out="sup").run(_)
            assert sys.stdout.getvalue() == "sup"

        @trap
        def err_defaults_to_sys_stderr(self):
            "err_stream defaults to sys.stderr"
            self._runner(err="sup").run(_)
            assert sys.stderr.getvalue() == "sup"

        @trap
        def out_can_be_overridden(self):
            "out_stream can be overridden"
            out = StringIO()
            self._runner(out="sup").run(_, out_stream=out)
            assert out.getvalue() == "sup"
            assert sys.stdout.getvalue() == ""

        @trap
        def overridden_out_is_never_hidden(self):
            out = StringIO()
            self._runner(out="sup").run(_, out_stream=out, hide=True)
            assert out.getvalue() == "sup"
            assert sys.stdout.getvalue() == ""

        @trap
        def err_can_be_overridden(self):
            "err_stream can be overridden"
            err = StringIO()
            self._runner(err="sup").run(_, err_stream=err)
            assert err.getvalue() == "sup"
            assert sys.stderr.getvalue() == ""

        @trap
        def overridden_err_is_never_hidden(self):
            err = StringIO()
            self._runner(err="sup").run(_, err_stream=err, hide=True)
            assert err.getvalue() == "sup"
            assert sys.stderr.getvalue() == ""

        @trap
        def pty_defaults_to_sys(self):
            self._runner(out="sup").run(_, pty=True)
            assert sys.stdout.getvalue() == "sup"

        @trap
        def pty_out_can_be_overridden(self):
            out = StringIO()
            self._runner(out="yo").run(_, pty=True, out_stream=out)
            assert out.getvalue() == "yo"
            assert sys.stdout.getvalue() == ""

    class output_stream_handling:
        # Mostly corner cases, generic behavior's covered above
        def writes_and_flushes_to_stdout(self):
            out = Mock(spec=StringIO)
            self._runner(out="meh").run(_, out_stream=out)
            out.write.assert_called_once_with("meh")
            out.flush.assert_called_once_with()

        def writes_and_flushes_to_stderr(self):
            err = Mock(spec=StringIO)
            self._runner(err="whatever").run(_, err_stream=err)
            err.write.assert_called_once_with("whatever")
            err.flush.assert_called_once_with()

    class input_stream_handling:
        # NOTE: actual autoresponder tests are elsewhere. These just test that
        # stdin works normally & can be overridden.
        @patch("invoke.runners.sys.stdin", StringIO("Text!"))
        def defaults_to_sys_stdin(self):
            # Execute w/ runner class that has a mocked stdin_writer
            klass = self._mock_stdin_writer()
            self._runner(klass=klass).run(_, out_stream=StringIO())
            # Check that mocked writer was called w/ the data from our patched
            # sys.stdin.
            # NOTE: this also tests that non-fileno-bearing streams read/write
            # 1 byte at a time. See farther-down test for fileno-bearing stdin
            calls = list(map(lambda x: call(x), "Text!"))
            klass.write_proc_stdin.assert_has_calls(calls, any_order=False)

        def can_be_overridden(self):
            klass = self._mock_stdin_writer()
            in_stream = StringIO("Hey, listen!")
            self._runner(klass=klass).run(
                _, in_stream=in_stream, out_stream=StringIO()
            )
            # stdin mirroring occurs char-by-char
            calls = list(map(lambda x: call(x), "Hey, listen!"))
            klass.write_proc_stdin.assert_has_calls(calls, any_order=False)

        def can_be_disabled_entirely(self):
            # Mock handle_stdin so we can assert it's not even called
            class MockedHandleStdin(_Dummy):
                pass

            MockedHandleStdin.handle_stdin = Mock()
            self._runner(klass=MockedHandleStdin).run(
                _, in_stream=False  # vs None or a stream
            )
            assert not MockedHandleStdin.handle_stdin.called

        @patch("invoke.util.debug")
        def exceptions_get_logged(self, mock_debug):
            # Make write_proc_stdin asplode
            klass = self._mock_stdin_writer()
            klass.write_proc_stdin.side_effect = OhNoz("oh god why")
            # Execute with some stdin to trigger that asplode (but skip the
            # actual bubbled-up raising of it so we can check things out)
            try:
                stdin = StringIO("non-empty")
                self._runner(klass=klass).run(_, in_stream=stdin)
            except ThreadException:
                pass
            # Assert debug() was called w/ expected format
            # TODO: make the debug call a method on ExceptionHandlingThread,
            # then make thread class configurable somewhere in Runner, and pass
            # in a customized ExceptionHandlingThread that has a Mock for that
            # method?
            # NOTE: splitting into a few asserts to work around python 3.7
            # change re: trailing comma, which kills ability to just statically
            # assert the entire string. Sigh. Also I'm too lazy to regex.
            msg = mock_debug.call_args[0][0]
            assert "Encountered exception OhNoz" in msg
            assert "'oh god why'" in msg
            assert "in thread for 'handle_stdin'" in msg

        def EOF_triggers_closing_of_proc_stdin(self):
            class Fake(_Dummy):
                pass

            Fake.close_proc_stdin = Mock()
            self._runner(klass=Fake).run(_, in_stream=StringIO("what?"))
            Fake.close_proc_stdin.assert_called_once_with()

        def EOF_does_not_close_proc_stdin_when_pty_True(self):
            class Fake(_Dummy):
                pass

            Fake.close_proc_stdin = Mock()
            self._runner(klass=Fake).run(
                _, in_stream=StringIO("what?"), pty=True
            )
            assert not Fake.close_proc_stdin.called

        @patch("invoke.runners.sys.stdin")
        def EBADF_on_stdin_read_ignored(self, fake_stdin):
            # Issue #659: nohup is a jerk
            fake_stdin.read.side_effect = OSError(errno.EBADF, "Ugh")
            # No boom == fixed
            self._runner().run(_)

        @patch("invoke.runners.sys.stdin")
        def non_EBADF_on_stdin_read_not_ignored(self, fake_stdin):
            # Issue #659: nohup is a jerk, inverse case
            eio = OSError(errno.EIO, "lol")
            fake_stdin.read.side_effect = eio
            with raises(ThreadException) as info:
                self._runner().run(_)
            assert info.value.exceptions[0].value is eio

    class failure_handling:
        def fast_failures(self):
            with raises(UnexpectedExit):
                self._runner(exits=1).run(_)

        def non_1_return_codes_still_act_as_failure(self):
            r = self._runner(exits=17).run(_, warn=True)
            assert r.failed is True

        class Failure_repr:
            def defaults_to_just_class_and_command(self):
                expected = "<Failure: cmd='oh hecc'>"
                assert repr(Failure(Result(command="oh hecc"))) == expected

            def subclasses_may_add_more_kv_pairs(self):
                class TotalFailure(Failure):
                    def _repr(self, **kwargs):
                        return super(TotalFailure, self)._repr(mood="dejected")

                expected = "<TotalFailure: cmd='onoz' mood=dejected>"
                assert repr(TotalFailure(Result(command="onoz"))) == expected

        class UnexpectedExit_repr:
            def similar_to_just_the_result_repr(self):
                try:
                    self._runner(exits=23).run(_)
                except UnexpectedExit as e:
                    expected = "<UnexpectedExit: cmd='{}' exited=23>"
                    assert repr(e) == expected.format(_)

        class UnexpectedExit_str:
            def setup_method(self):
                def lines(prefix):
                    prefixed = "\n".join(
                        "{} {}".format(prefix, x) for x in range(1, 26)
                    )
                    return prefixed + "\n"

                self._stdout = lines("stdout")
                self._stderr = lines("stderr")

            @trap
            def displays_command_and_exit_code_by_default(self):
                try:
                    self._runner(
                        exits=23, out=self._stdout, err=self._stderr
                    ).run(_)
                except UnexpectedExit as e:
                    expected = """Encountered a bad command exit code!

Command: '{}'

Exit code: 23

Stdout: already printed

Stderr: already printed

"""
                    assert str(e) == expected.format(_)
                else:
                    assert False, "Failed to raise UnexpectedExit!"

            @trap
            def does_not_display_stderr_when_pty_True(self):
                try:
                    self._runner(
                        exits=13, out=self._stdout, err=self._stderr
                    ).run(_, pty=True)
                except UnexpectedExit as e:
                    expected = """Encountered a bad command exit code!

Command: '{}'

Exit code: 13

Stdout: already printed

Stderr: n/a (PTYs have no stderr)

"""
                    assert str(e) == expected.format(_)

            @trap
            def pty_stderr_message_wins_over_hidden_stderr(self):
                try:
                    self._runner(
                        exits=1, out=self._stdout, err=self._stderr
                    ).run(_, pty=True, hide=True)
                except UnexpectedExit as e:
                    r = str(e)
                    assert "Stderr: n/a (PTYs have no stderr)" in r
                    assert "Stderr: already printed" not in r

            @trap
            def explicit_hidden_stream_tail_display(self):
                # All the permutations of what's displayed when, are in
                # subsequent test, which does 'x in y' assertions; this one
                # here ensures the actual format of the display (newlines, etc)
                # is as desired.
                try:
                    self._runner(
                        exits=77, out=self._stdout, err=self._stderr
                    ).run(_, hide=True)
                except UnexpectedExit as e:
                    expected = """Encountered a bad command exit code!

Command: '{}'

Exit code: 77

Stdout:

stdout 16
stdout 17
stdout 18
stdout 19
stdout 20
stdout 21
stdout 22
stdout 23
stdout 24
stdout 25

Stderr:

stderr 16
stderr 17
stderr 18
stderr 19
stderr 20
stderr 21
stderr 22
stderr 23
stderr 24
stderr 25

"""
                    assert str(e) == expected.format(_)

            @trap
            def displays_tails_of_streams_only_when_hidden(self):
                def oops(msg, r, hide):
                    return "{}! hide={}; str output:\n\n{}".format(
                        msg, hide, r
                    )

                for hide, expect_out, expect_err in (
                    (False, False, False),
                    (True, True, True),
                    ("stdout", True, False),
                    ("stderr", False, True),
                    ("both", True, True),
                ):
                    try:
                        self._runner(
                            exits=1, out=self._stdout, err=self._stderr
                        ).run(_, hide=hide)
                    except UnexpectedExit as e:
                        r = str(e)
                        # Expect that the top of output is never displayed
                        err = oops("Too much stdout found", r, hide)
                        assert "stdout 15" not in r, err
                        err = oops("Too much stderr found", r, hide)
                        assert "stderr 15" not in r, err
                        # Expect to see tail of stdout if we expected it
                        if expect_out:
                            err = oops("Didn't see stdout", r, hide)
                            assert "stdout 16" in r, err
                        # Expect to see tail of stderr if we expected it
                        if expect_err:
                            err = oops("Didn't see stderr", r, hide)
                            assert "stderr 16" in r, err
                    else:
                        assert False, "Failed to raise UnexpectedExit!"

        def _regular_error(self):
            self._runner(exits=1).run(_)

        def _watcher_error(self):
            klass = self._mock_stdin_writer()
            # Exited=None because real procs will have no useful .returncode()
            # result if they're aborted partway via an exception.
            runner = self._runner(klass=klass, out="stuff", exits=None)
            runner.run(_, watchers=[_RaisingWatcher()], hide=True)

        # TODO: may eventually turn into having Runner raise distinct Failure
        # subclasses itself, at which point `reason` would probably go away.
        class reason:
            def is_None_for_regular_nonzero_exits(self):
                try:
                    self._regular_error()
                except Failure as e:
                    assert e.reason is None
                else:
                    assert False, "Failed to raise Failure!"

            def is_None_for_custom_command_exits(self):
                # TODO: when we implement 'exitcodes 1 and 2 are actually OK'
                skip()

            def is_exception_when_WatcherError_raised_internally(self):
                try:
                    self._watcher_error()
                except Failure as e:
                    assert isinstance(e.reason, WatcherError)
                else:
                    assert False, "Failed to raise Failure!"

        # TODO: should these move elsewhere, eg to Result specific test file?
        # TODO: *is* there a nice way to split into multiple Response and/or
        # Failure subclasses? Given the split between "returned as a value when
        # no problem" and "raised as/attached to an exception when problem",
        # possibly not - complicates how the APIs need to be adhered to.
        class wrapped_result:
            def most_attrs_are_always_present(self):
                attrs = ("command", "shell", "env", "stdout", "stderr", "pty")
                for method in (self._regular_error, self._watcher_error):
                    try:
                        method()
                    except Failure as e:
                        for attr in attrs:
                            assert getattr(e.result, attr) is not None
                    else:
                        assert False, "Did not raise Failure!"

            class shell_exit_failure:
                def exited_is_integer(self):
                    try:
                        self._regular_error()
                    except Failure as e:
                        assert isinstance(e.result.exited, int)
                    else:
                        assert False, "Did not raise Failure!"

                def ok_bool_etc_are_falsey(self):
                    try:
                        self._regular_error()
                    except Failure as e:
                        assert e.result.ok is False
                        assert e.result.failed is True
                        assert not bool(e.result)
                        assert not e.result
                    else:
                        assert False, "Did not raise Failure!"

                def stringrep_notes_exit_status(self):
                    try:
                        self._regular_error()
                    except Failure as e:
                        assert "exited with status 1" in str(e.result)
                    else:
                        assert False, "Did not raise Failure!"

            class watcher_failure:
                def exited_is_None(self):
                    try:
                        self._watcher_error()
                    except Failure as e:
                        exited = e.result.exited
                        err = "Expected None, got {!r}".format(exited)
                        assert exited is None, err

                def ok_and_bool_still_are_falsey(self):
                    try:
                        self._watcher_error()
                    except Failure as e:
                        assert e.result.ok is False
                        assert e.result.failed is True
                        assert not bool(e.result)
                        assert not e.result
                    else:
                        assert False, "Did not raise Failure!"

                def stringrep_lacks_exit_status(self):
                    try:
                        self._watcher_error()
                    except Failure as e:
                        assert "exited with status" not in str(e.result)
                        expected = "not fully executed due to watcher error"
                        assert expected in str(e.result)
                    else:
                        assert False, "Did not raise Failure!"

    class threading:
        # NOTE: see also the more generic tests in concurrency.py
        def errors_within_io_thread_body_bubble_up(self):
            class Oops(_Dummy):
                def handle_stdout(self, **kwargs):
                    raise OhNoz()

                def handle_stderr(self, **kwargs):
                    raise OhNoz()

            runner = Oops(Context())
            try:
                runner.run("nah")
            except ThreadException as e:
                # Expect two separate OhNoz objects on 'e'
                assert len(e.exceptions) == 2
                for tup in e.exceptions:
                    assert isinstance(tup.value, OhNoz)
                    assert isinstance(tup.traceback, types.TracebackType)
                    assert tup.type == OhNoz
                # TODO: test the arguments part of the tuple too. It's pretty
                # implementation-specific, though, so possibly not worthwhile.
            else:
                assert False, "Did not raise ThreadException as expected!"

        def io_thread_errors_str_has_details(self):
            class Oops(_Dummy):
                def handle_stdout(self, **kwargs):
                    raise OhNoz()

            runner = Oops(Context())
            try:
                runner.run("nah")
            except ThreadException as e:
                message = str(e)
                # Just make sure salient bits appear present, vs e.g. default
                # representation happening instead.
                assert "Saw 1 exceptions within threads" in message
                assert "{'kwargs': " in message
                assert "Traceback (most recent call last):\n\n" in message
                assert "OhNoz" in message
            else:
                assert False, "Did not raise ThreadException as expected!"

    class watchers:
        # NOTE: it's initially tempting to consider using mocks or stub
        # Responder instances for many of these, but it really doesn't save
        # appreciable runtime or code read/write time.
        # NOTE: these strictly test interactions between
        # StreamWatcher/Responder and their host Runner; Responder-only tests
        # are in tests/watchers.py.

        def nothing_is_written_to_stdin_by_default(self):
            # NOTE: technically if some goofus ran the tests by hand and mashed
            # keys while doing so...this would fail. LOL?
            # NOTE: this test seems not too useful but is a) a sanity test and
            # b) guards against e.g. breaking the autoresponder such that it
            # responds to "" or "\n" or etc.
            klass = self._mock_stdin_writer()
            self._runner(klass=klass).run(_)
            assert not klass.write_proc_stdin.called

        def _expect_response(self, **kwargs):
            """
            Execute a run() w/ ``watchers`` set from ``responses``.

            Any other ``**kwargs`` given are passed direct to ``_runner()``.

            :returns: The mocked ``write_proc_stdin`` method of the runner.
            """
            watchers = [
                Responder(pattern=key, response=value)
                for key, value in kwargs.pop("responses").items()
            ]
            kwargs["klass"] = klass = self._mock_stdin_writer()
            runner = self._runner(**kwargs)
            runner.run(_, watchers=watchers, hide=True)
            return klass.write_proc_stdin

        def watchers_responses_get_written_to_proc_stdin(self):
            self._expect_response(
                out="the house was empty", responses={"empty": "handed"}
            ).assert_called_once_with("handed")

        def multiple_hits_yields_multiple_responses(self):
            holla = call("how high?")
            self._expect_response(
                out="jump, wait, jump, wait", responses={"jump": "how high?"}
            ).assert_has_calls([holla, holla])

        def chunk_sizes_smaller_than_patterns_still_work_ok(self):
            klass = self._mock_stdin_writer()
            klass.read_chunk_size = 1  # < len('jump')
            responder = Responder("jump", "how high?")
            runner = self._runner(klass=klass, out="jump, wait, jump, wait")
            runner.run(_, watchers=[responder], hide=True)
            holla = call("how high?")
            # Responses happened, period.
            klass.write_proc_stdin.assert_has_calls([holla, holla])
            # And there weren't duplicates!
            assert len(klass.write_proc_stdin.call_args_list) == 2

        def both_out_and_err_are_scanned(self):
            bye = call("goodbye")
            # Would only be one 'bye' if only scanning stdout
            self._expect_response(
                out="hello my name is inigo",
                err="hello how are you",
                responses={"hello": "goodbye"},
            ).assert_has_calls([bye, bye])

        def multiple_patterns_works_as_expected(self):
            calls = [call("betty"), call("carnival")]
            self._expect_response(
                out="beep boop I am a robot",
                responses={"boop": "betty", "robot": "carnival"},
            ).assert_has_calls(calls, any_order=True)

        def multiple_patterns_across_both_streams(self):
            responses = {
                "boop": "betty",
                "robot": "carnival",
                "Destroy": "your ego",
                "humans": "are awful",
            }
            calls = map(lambda x: call(x), responses.values())
            # CANNOT assume order due to simultaneous streams.
            # If we didn't say any_order=True we could get race condition fails
            self._expect_response(
                out="beep boop, I am a robot",
                err="Destroy all humans!",
                responses=responses,
            ).assert_has_calls(calls, any_order=True)

        def honors_watchers_config_option(self):
            klass = self._mock_stdin_writer()
            responder = Responder("my stdout", "and my axe")
            runner = self._runner(
                out="this is my stdout",  # yielded stdout
                klass=klass,  # mocked stdin writer
                run={"watchers": [responder]},  # ends up as config override
            )
            runner.run(_, hide=True)
            klass.write_proc_stdin.assert_called_once_with("and my axe")

        def kwarg_overrides_config(self):
            # TODO: how to handle use cases where merging, not overriding, is
            # the expected/unsurprising default? probably another config-only
            # (not kwarg) setting, e.g. run.merge_responses?
            # TODO: now that this stuff is list, not dict, based, it should be
            # easier...BUT how to handle removal of defaults from config? Maybe
            # just document to be careful using the config as it won't _be_
            # overridden? (Users can always explicitly set the config to be
            # empty-list if they want kwargs to be the entire set of
            # watchers...right?)
            klass = self._mock_stdin_writer()
            conf = Responder("my stdout", "and my axe")
            kwarg = Responder("my stdout", "and my body spray")
            runner = self._runner(
                out="this is my stdout",  # yielded stdout
                klass=klass,  # mocked stdin writer
                run={"watchers": [conf]},  # ends up as config override
            )
            runner.run(_, hide=True, watchers=[kwarg])
            klass.write_proc_stdin.assert_called_once_with("and my body spray")

    class io_sleeping:
        # NOTE: there's an explicit CPU-measuring test in the integration suite
        # which ensures the *point* of the sleeping - avoiding CPU hogging - is
        # actually functioning. These tests below just unit-test the mechanisms
        # around the sleep functionality (ensuring they are visible and can be
        # altered as needed).
        def input_sleep_attribute_defaults_to_hundredth_of_second(self):
            assert Runner(Context()).input_sleep == 0.01

        @mock_subprocess()
        def subclasses_can_override_input_sleep(self):
            class MyRunner(_Dummy):
                input_sleep = 0.007

            with patch("invoke.runners.time") as mock_time:
                MyRunner(Context()).run(
                    _,
                    in_stream=StringIO("foo"),
                    out_stream=StringIO(),  # null output to not pollute tests
                )
            # Just make sure the first few sleeps all look good. Can't know
            # exact length of list due to stdin worker hanging out til end of
            # process. Still worth testing more than the first tho.
            assert mock_time.sleep.call_args_list[:3] == [call(0.007)] * 3

    class stdin_mirroring:
        def _test_mirroring(self, expect_mirroring, **kwargs):
            # Setup
            fake_in = "I'm typing!"
            output = Mock()
            input_ = StringIO(fake_in)
            input_is_pty = kwargs.pop("in_pty", None)

            class MyRunner(_Dummy):
                def should_echo_stdin(self, input_, output):
                    # Fake result of isatty() test here and only here; if we do
                    # this farther up, it will affect stuff trying to run
                    # termios & such, which is harder to mock successfully.
                    if input_is_pty is not None:
                        input_.isatty = lambda: input_is_pty
                    return super(MyRunner, self).should_echo_stdin(
                        input_, output
                    )

            # Execute basic command with given parameters
            self._run(
                _,
                klass=MyRunner,
                in_stream=input_,
                out_stream=output,
                **kwargs
            )
            # Examine mocked output stream to see if it was mirrored to
            if expect_mirroring:
                calls = output.write.call_args_list
                assert calls == list(map(lambda x: call(x), fake_in))
                assert len(output.flush.call_args_list) == len(fake_in)
            # Or not mirrored to
            else:
                assert output.write.call_args_list == []

        def when_pty_is_True_no_mirroring_occurs(self):
            self._test_mirroring(pty=True, expect_mirroring=False)

        def when_pty_is_False_we_write_in_stream_back_to_out_stream(self):
            self._test_mirroring(pty=False, in_pty=True, expect_mirroring=True)

        def mirroring_is_skipped_when_our_input_is_not_a_tty(self):
            self._test_mirroring(in_pty=False, expect_mirroring=False)

        def mirroring_can_be_forced_on(self):
            self._test_mirroring(
                # Subprocess pty normally disables echoing
                pty=True,
                # But then we forcibly enable it
                echo_stdin=True,
                # And expect it to happen
                expect_mirroring=True,
            )

        def mirroring_can_be_forced_off(self):
            # Make subprocess pty False, stdin tty True, echo_stdin False,
            # prove no mirroring
            self._test_mirroring(
                # Subprocess lack of pty normally enables echoing
                pty=False,
                # Provided the controlling terminal _is_ a tty
                in_pty=True,
                # But then we forcibly disable it
                echo_stdin=False,
                # And expect it to not happen
                expect_mirroring=False,
            )

        def mirroring_honors_configuration(self):
            self._test_mirroring(
                pty=False,
                in_pty=True,
                settings={"run": {"echo_stdin": False}},
                expect_mirroring=False,
            )

        @trap
        @skip_if_windows
        @patch("invoke.runners.sys.stdin")
        @patch("invoke.terminals.fcntl.ioctl")
        @patch("invoke.terminals.os")
        @patch("invoke.terminals.termios")
        @patch("invoke.terminals.tty")
        @patch("invoke.terminals.select")
        # NOTE: the no-fileno edition is handled at top of this local test
        # class, in the base case test.
        def reads_FIONREAD_bytes_from_stdin_when_fileno(
            self, select, tty, termios, mock_os, ioctl, stdin
        ):
            # Set stdin up as a file-like buffer which passes has_fileno
            stdin.fileno.return_value = 17  # arbitrary
            stdin_data = list("boo!")

            def fakeread(n):
                # Why is there no slice version of pop()?
                data = stdin_data[:n]
                del stdin_data[:n]
                return "".join(data)

            stdin.read.side_effect = fakeread
            # Without mocking this, we'll always get errors checking the above
            # bogus fileno()
            mock_os.tcgetpgrp.return_value = None
            # Ensure select() only spits back stdin one time, despite there
            # being multiple bytes to read (this at least partly fakes behavior
            # from issue #58)
            select.select.side_effect = chain(
                [([stdin], [], [])], repeat(([], [], []))
            )
            # Have ioctl yield our multiple number of bytes when called with
            # FIONREAD
            def fake_ioctl(fd, cmd, buf):
                # This works since each mocked attr will still be its own mock
                # object with a distinct 'is' identity.
                if cmd is termios.FIONREAD:
                    return struct.pack("h", len(stdin_data))

            ioctl.side_effect = fake_ioctl
            # Set up our runner as one w/ mocked stdin writing (simplest way to
            # assert how the reads & writes are happening)
            klass = self._mock_stdin_writer()
            self._runner(klass=klass).run(_)
            klass.write_proc_stdin.assert_called_once_with("boo!")

    class character_buffered_stdin:
        @skip_if_windows
        @patch("invoke.terminals.tty")
        def setcbreak_called_on_tty_stdins(self, mock_tty, mock_termios):
            mock_termios.tcgetattr.return_value = _make_tcattrs(echo=True)
            self._run(_)
            mock_tty.setcbreak.assert_called_with(sys.stdin)

        @skip_if_windows
        @patch("invoke.terminals.tty")
        def setcbreak_not_called_on_non_tty_stdins(self, mock_tty):
            self._run(_, in_stream=StringIO())
            assert not mock_tty.setcbreak.called

        @skip_if_windows
        @patch("invoke.terminals.tty")
        @patch("invoke.terminals.os")
        def setcbreak_not_called_if_process_not_foregrounded(
            self, mock_os, mock_tty
        ):
            # Re issue #439.
            mock_os.getpgrp.return_value = 1337
            mock_os.tcgetpgrp.return_value = 1338
            self._run(_)
            assert not mock_tty.setcbreak.called
            # Sanity
            mock_os.tcgetpgrp.assert_called_once_with(sys.stdin.fileno())

        @skip_if_windows
        @patch("invoke.terminals.tty")
        def tty_stdins_have_settings_restored_by_default(
            self, mock_tty, mock_termios
        ):
            # Get already-cbroken attrs since that's an easy way to get the
            # right format/layout
            attrs = _make_tcattrs(echo=True)
            mock_termios.tcgetattr.return_value = attrs
            self._run(_)
            # Ensure those old settings are being restored
            mock_termios.tcsetattr.assert_called_once_with(
                sys.stdin, mock_termios.TCSADRAIN, attrs
            )

        @skip_if_windows
        @patch("invoke.terminals.tty")  # stub
        def tty_stdins_have_settings_restored_on_KeyboardInterrupt(
            self, mock_tty, mock_termios
        ):
            # This test is re: GH issue #303
            sentinel = _make_tcattrs(echo=True)
            mock_termios.tcgetattr.return_value = sentinel
            # Don't actually bubble up the KeyboardInterrupt...
            try:
                self._run(_, klass=_KeyboardInterruptingRunner)
            except KeyboardInterrupt:
                pass
            # Did we restore settings?!
            mock_termios.tcsetattr.assert_called_once_with(
                sys.stdin, mock_termios.TCSADRAIN, sentinel
            )

        @skip_if_windows
        @patch("invoke.terminals.tty")
        def setcbreak_not_called_if_terminal_seems_already_cbroken(
            self, mock_tty, mock_termios
        ):
            # Proves #559, sorta, insofar as it only passes when the fixed
            # behavior is in place. (Proving the old bug is hard as it is race
            # condition reliant; the new behavior sidesteps that entirely.)

            # Test both bytes and ints versions of CC values, since docs
            # disagree with at least some platforms' realities on that.
            for is_ints in (True, False):
                mock_termios.tcgetattr.return_value = _make_tcattrs(
                    cc_is_ints=is_ints
                )
                self._run(_)
                # Ensure tcsetattr and setcbreak were never called
                assert not mock_tty.setcbreak.called
                assert not mock_termios.tcsetattr.called

    class send_interrupt:
        def _run_with_mocked_interrupt(self, klass):
            runner = klass(Context())
            runner.send_interrupt = Mock()
            try:
                runner.run(_)
            except _GenericException:
                pass
            return runner

        def called_on_KeyboardInterrupt(self):
            runner = self._run_with_mocked_interrupt(
                _KeyboardInterruptingRunner
            )
            assert runner.send_interrupt.called

        def not_called_for_other_exceptions(self):
            runner = self._run_with_mocked_interrupt(_GenericExceptingRunner)
            assert not runner.send_interrupt.called

        def sends_escape_byte_sequence(self):
            for pty in (True, False):
                runner = _KeyboardInterruptingRunner(Context())
                mock_stdin = Mock()
                runner.write_proc_stdin = mock_stdin
                runner.run(_, pty=pty)
                mock_stdin.assert_called_once_with("\x03")

    class timeout:
        def start_timer_called_with_config_value(self):
            runner = self._runner(timeouts={"command": 7})
            runner.start_timer = Mock()
            assert runner.context.config.timeouts.command == 7
            runner.run(_)
            runner.start_timer.assert_called_once_with(7)

        def run_kwarg_honored(self):
            runner = self._runner()
            runner.start_timer = Mock()
            assert runner.context.config.timeouts.command is None
            runner.run(_, timeout=3)
            runner.start_timer.assert_called_once_with(3)

        def kwarg_wins_over_config(self):
            runner = self._runner(timeouts={"command": 7})
            runner.start_timer = Mock()
            assert runner.context.config.timeouts.command == 7
            runner.run(_, timeout=3)
            runner.start_timer.assert_called_once_with(3)

        def raises_CommandTimedOut_with_timeout_info(self):
            runner = self._runner(
                klass=_TimingOutRunner, timeouts={"command": 7}
            )
            with raises(CommandTimedOut) as info:
                runner.run(_)
            assert info.value.timeout == 7
            _repr = "<CommandTimedOut: cmd='nope' timeout=7>"
            assert repr(info.value) == _repr
            expected = """
Command did not complete within 7 seconds!

Command: 'nope'

Stdout: already printed

Stderr: already printed

""".lstrip()
            assert str(info.value) == expected

        @patch("invoke.runners.threading.Timer")
        def start_timer_gives_its_timer_the_kill_method(self, Timer):
            runner = self._runner()
            runner.start_timer(30)
            Timer.assert_called_once_with(30, runner.kill)

        def _mocked_timer(self):
            runner = self._runner()
            runner._timer = Mock()
            return runner

        def run_always_stops_timer(self):
            runner = _GenericExceptingRunner(Context())
            runner._timer = Mock()
            with raises(_GenericException):
                runner.run(_)
            runner._timer.cancel.assert_called_once_with()

        def stop_cancels_timer(self):
            runner = self._mocked_timer()
            runner.stop()
            runner._timer.cancel.assert_called_once_with()

        def timer_aliveness_is_test_of_timing_out(self):
            # Might be redundant, but easy enough to unit test
            runner = Runner(Context())
            runner._timer = Mock()
            runner._timer.is_alive.return_value = False
            assert runner.timed_out
            runner._timer.is_alive.return_value = True
            assert not runner.timed_out

        def timeout_specified_but_no_timer_means_no_exception(self):
            # Weird corner case but worth testing
            runner = Runner(Context())
            runner._timer = None
            assert not runner.timed_out

    class stop:
        def always_runs_no_matter_what(self):
            runner = _GenericExceptingRunner(context=Context())
            runner.stop = Mock()
            with raises(_GenericException):
                runner.run(_)
            runner.stop.assert_called_once_with()

    class asynchronous:
        def returns_Promise_immediately_and_finishes_on_join(self):
            # Dummy subclass with controllable process_is_finished flag
            class _Finisher(_Dummy):
                _finished = False

                @property
                def process_is_finished(self):
                    return self._finished

            runner = _Finisher(Context())
            # Set up mocks and go
            runner.start = Mock()
            for method in self._stop_methods:
                setattr(runner, method, Mock())
            result = runner.run(_, asynchronous=True)
            # Got a Promise (its attrs etc are in its own test subsuite)
            assert isinstance(result, Promise)
            # Started, but did not stop (as would've happened for disown)
            assert runner.start.called
            for method in self._stop_methods:
                assert not getattr(runner, method).called
            # Set proc completion flag to truthy and join()
            runner._finished = True
            result.join()
            for method in self._stop_methods:
                assert getattr(runner, method).called

        @trap
        def hides_output(self):
            # Run w/ faux subproc stdout/err data, but async
            self._runner(out="foo", err="bar").run(_, asynchronous=True).join()
            # Expect that default out/err streams did not get printed to.
            assert sys.stdout.getvalue() == ""
            assert sys.stderr.getvalue() == ""

        def does_not_forward_stdin(self):
            class MockedHandleStdin(_Dummy):
                pass

            MockedHandleStdin.handle_stdin = Mock()
            runner = self._runner(klass=MockedHandleStdin)
            runner.run(_, asynchronous=True).join()
            # As with the main test for setting this to False, we know that
            # when stdin is disabled, the handler is never even called (no
            # thread is created for it).
            assert not MockedHandleStdin.handle_stdin.called

        def leaves_overridden_streams_alone(self):
            # NOTE: technically a duplicate test of the generic tests for #637
            # re: intersect of hide and overridden streams. But that's an
            # implementation detail so this is still valuable.
            klass = self._mock_stdin_writer()
            out, err, in_ = StringIO(), StringIO(), StringIO("hallo")
            runner = self._runner(out="foo", err="bar", klass=klass)
            runner.run(
                _,
                asynchronous=True,
                out_stream=out,
                err_stream=err,
                in_stream=in_,
            ).join()
            assert out.getvalue() == "foo"
            assert err.getvalue() == "bar"
            assert klass.write_proc_stdin.called  # lazy

    class disown:
        @patch.object(threading.Thread, "start")
        def starts_and_returns_None_but_does_nothing_else(self, thread_start):
            runner = Runner(Context())
            runner.start = Mock()
            not_called = self._stop_methods + ["wait"]
            for method in not_called:
                setattr(runner, method, Mock())
            result = runner.run(_, disown=True)
            # No Result object!
            assert result is None
            # Subprocess kicked off
            assert runner.start.called
            # No timer or IO threads started
            assert not thread_start.called
            # No wait or shutdown related Runner methods called
            for method in not_called:
                assert not getattr(runner, method).called

        def cannot_be_given_alongside_asynchronous(self):
            with raises(ValueError) as info:
                self._runner().run(_, asynchronous=True, disown=True)
            sentinel = "Cannot give both 'asynchronous' and 'disown'"
            assert sentinel in str(info.value)


class _FastLocal(Local):
    # Neuter this for same reason as in _Dummy above
    input_sleep = 0


class Local_:
    def _run(self, *args, **kwargs):
        return _run(*args, **dict(kwargs, klass=_FastLocal))

    def _runner(self, *args, **kwargs):
        return _runner(*args, **dict(kwargs, klass=_FastLocal))

    class pty:
        @mock_pty()
        def when_pty_True_we_use_pty_fork_and_os_exec(self):
            "when pty=True, we use pty.fork and os.exec*"
            self._run(_, pty=True)
            # @mock_pty's asserts check os/pty calls for us.

        @mock_pty(insert_os=True)
        def _expect_exit_check(self, exited, mock_os):
            if exited:
                expected_check = mock_os.WIFEXITED
                expected_get = mock_os.WEXITSTATUS
                unexpected_check = mock_os.WIFSIGNALED
                unexpected_get = mock_os.WTERMSIG
            else:
                expected_check = mock_os.WIFSIGNALED
                expected_get = mock_os.WTERMSIG
                unexpected_check = mock_os.WIFEXITED
                unexpected_get = mock_os.WEXITSTATUS
            expected_check.return_value = True
            unexpected_check.return_value = False
            self._run(_, pty=True)
            exitstatus = mock_os.waitpid.return_value[1]
            expected_get.assert_called_once_with(exitstatus)
            assert not unexpected_get.called

        def pty_uses_WEXITSTATUS_if_WIFEXITED(self):
            self._expect_exit_check(True)

        def pty_uses_WTERMSIG_if_WIFSIGNALED(self):
            self._expect_exit_check(False)

        @mock_pty(insert_os=True)
        def WTERMSIG_result_turned_negative_to_match_subprocess(self, mock_os):
            mock_os.WIFEXITED.return_value = False
            mock_os.WIFSIGNALED.return_value = True
            mock_os.WTERMSIG.return_value = 2
            assert self._run(_, pty=True, warn=True).exited == -2

        @mock_pty()
        def pty_is_set_to_controlling_terminal_size(self):
            self._run(_, pty=True)
            # @mock_pty's asserts check the TIOC[GS]WINSZ calls for us

        def warning_only_fires_once(self):
            # I.e. if implementation checks pty-ness >1 time, only one warning
            # is emitted. This is kinda implementation-specific, but...
            skip()

        @patch("invoke.runners.sys")
        def replaced_stdin_objects_dont_explode(self, mock_sys):
            # Replace sys.stdin with an object lacking .isatty(), which
            # normally causes an AttributeError unless we are being careful.
            mock_sys.stdin = object()
            # Test. If bug is present, this will error.
            runner = Local(Context())
            assert runner.should_use_pty(pty=True, fallback=True) is False

        @mock_pty(trailing_error=OSError("Input/output error"))
        def spurious_OSErrors_handled_gracefully(self):
            # Doesn't-blow-up test.
            self._run(_, pty=True)

        @mock_pty(trailing_error=OSError("I/O error"))
        def other_spurious_OSErrors_handled_gracefully(self):
            # Doesn't-blow-up test.
            self._run(_, pty=True)

        @mock_pty(trailing_error=OSError("wat"))
        def non_spurious_OSErrors_bubble_up(self):
            try:
                self._run(_, pty=True)
            except ThreadException as e:
                e = e.exceptions[0]
                assert e.type == OSError
                assert str(e.value) == "wat"

        @mock_pty(os_close_error=True)
        def stop_mutes_errors_on_pty_close(self):
            # Another doesn't-blow-up test, this time around os.close() of the
            # pty itself (due to os_close_error=True)
            self._run(_, pty=True)

        class fallback:
            @mock_pty(isatty=False)
            def can_be_overridden_by_kwarg(self):
                self._run(_, pty=True, fallback=False)
                # @mock_pty's asserts will be mad if pty-related os/pty calls
                # didn't fire, so we're done.

            @mock_pty(isatty=False)
            def can_be_overridden_by_config(self):
                self._runner(run={"fallback": False}).run(_, pty=True)
                # @mock_pty's asserts will be mad if pty-related os/pty calls
                # didn't fire, so we're done.

            @trap
            @mock_subprocess(isatty=False)
            def affects_result_pty_value(self, *mocks):
                assert self._run(_, pty=True).pty is False

            @mock_pty(isatty=False)
            def overridden_fallback_affects_result_pty_value(self):
                assert self._run(_, pty=True, fallback=False).pty is True

    class shell:
        @mock_pty(insert_os=True)
        def defaults_to_bash_or_cmdexe_when_pty_True(self, mock_os):
            # NOTE: yea, windows can't run pty is true, but this is really
            # testing config behavior, so...meh
            self._run(_, pty=True)
            _expect_platform_shell(mock_os.execve.call_args_list[0][0][0])

        @mock_subprocess(insert_Popen=True)
        def defaults_to_bash_or_cmdexe_when_pty_False(self, mock_Popen):
            self._run(_, pty=False)
            _expect_platform_shell(
                mock_Popen.call_args_list[0][1]["executable"]
            )

        @mock_pty(insert_os=True)
        def may_be_overridden_when_pty_True(self, mock_os):
            self._run(_, pty=True, shell="/bin/zsh")
            assert mock_os.execve.call_args_list[0][0][0] == "/bin/zsh"

        @mock_subprocess(insert_Popen=True)
        def may_be_overridden_when_pty_False(self, mock_Popen):
            self._run(_, pty=False, shell="/bin/zsh")
            assert mock_Popen.call_args_list[0][1]["executable"] == "/bin/zsh"

    class env:
        # NOTE: update-vs-replace semantics are tested 'purely' up above in
        # regular Runner tests.

        @mock_subprocess(insert_Popen=True)
        def uses_Popen_kwarg_for_pty_False(self, mock_Popen):
            self._run(_, pty=False, env={"FOO": "BAR"})
            expected = dict(os.environ, FOO="BAR")
            env = mock_Popen.call_args_list[0][1]["env"]
            assert env == expected

        @mock_pty(insert_os=True)
        def uses_execve_for_pty_True(self, mock_os):
            type(mock_os).environ = {"OTHERVAR": "OTHERVAL"}
            self._run(_, pty=True, env={"FOO": "BAR"})
            expected = {"OTHERVAR": "OTHERVAL", "FOO": "BAR"}
            env = mock_os.execve.call_args_list[0][0][2]
            assert env == expected

    class close_proc_stdin:
        def raises_SubprocessPipeError_when_pty_in_use(self):
            with raises(SubprocessPipeError):
                runner = Local(Context())
                runner.using_pty = True
                runner.close_proc_stdin()

        def closes_process_stdin(self):
            runner = Local(Context())
            runner.process = Mock()
            runner.using_pty = False
            runner.close_proc_stdin()
            runner.process.stdin.close.assert_called_once_with()

    class timeout:
        @patch("invoke.runners.os")
        def kill_uses_self_pid_when_pty(self, mock_os):
            runner = self._runner()
            runner.using_pty = True
            runner.pid = 50
            runner.kill()
            mock_os.kill.assert_called_once_with(50, signal.SIGKILL)

        @patch("invoke.runners.os")
        def kill_uses_self_process_pid_when_not_pty(self, mock_os):
            runner = self._runner()
            runner.using_pty = False
            runner.process = Mock(pid=30)
            runner.kill()
            mock_os.kill.assert_called_once_with(30, signal.SIGKILL)


class Result_:
    def nothing_is_required(self):
        Result()

    def first_posarg_is_stdout(self):
        assert Result("foo").stdout == "foo"

    def command_defaults_to_empty_string(self):
        assert Result().command == ""

    def shell_defaults_to_empty_string(self):
        assert Result().shell == ""

    def encoding_defaults_to_local_default_encoding(self):
        assert Result().encoding == default_encoding()

    def env_defaults_to_empty_dict(self):
        assert Result().env == {}

    def stdout_defaults_to_empty_string(self):
        assert Result().stdout == ""

    def stderr_defaults_to_empty_string(self):
        assert Result().stderr == ""

    def exited_defaults_to_zero(self):
        assert Result().exited == 0

    def pty_defaults_to_False(self):
        assert Result().pty is False

    def repr_contains_useful_info(self):
        assert repr(Result(command="foo")) == "<Result cmd='foo' exited=0>"

    class tail:
        def setup_method(self):
            self.sample = "\n".join(str(x) for x in range(25))

        def returns_last_10_lines_of_given_stream_plus_whitespace(self):
            expected = """

15
16
17
18
19
20
21
22
23
24"""
            assert Result(stdout=self.sample).tail("stdout") == expected

        def line_count_is_configurable(self):
            expected = """

23
24"""
            tail = Result(stdout=self.sample).tail("stdout", count=2)
            assert tail == expected

        def works_for_stderr_too(self):
            # Dumb test is dumb, but whatever
            expected = """

23
24"""
            tail = Result(stderr=self.sample).tail("stderr", count=2)
            assert tail == expected


class Promise_:
    def exposes_read_only_run_params(self):
        runner = _runner()
        promise = runner.run(
            _, pty=True, encoding="utf-17", shell="sea", asynchronous=True
        )
        assert promise.command == _
        assert promise.pty is True
        assert promise.encoding == "utf-17"
        assert promise.shell == "sea"
        assert not hasattr(promise, "stdout")
        assert not hasattr(promise, "stderr")

    class join:
        # NOTE: high level Runner lifecycle mechanics of join() (re: wait(),
        # process_is_finished() etc) are tested in main suite.

        def returns_Result_on_success(self):
            result = _runner().run(_, asynchronous=True).join()
            assert isinstance(result, Result)
            # Sanity
            assert result.command == _
            assert result.exited == 0

        def raises_main_thread_exception_on_kaboom(self):
            runner = _runner(klass=_GenericExceptingRunner)
            with raises(_GenericException):
                runner.run(_, asynchronous=True).join()

        def raises_subthread_exception_on_their_kaboom(self):
            class Kaboom(_Dummy):
                def handle_stdout(self, **kwargs):
                    raise OhNoz()

            runner = _runner(klass=Kaboom)
            promise = runner.run(_, asynchronous=True)
            with raises(ThreadException) as info:
                promise.join()
            assert isinstance(info.value.exceptions[0].value, OhNoz)

        def raises_Failure_on_failure(self):
            runner = _runner(exits=1)
            promise = runner.run(_, asynchronous=True)
            with raises(Failure):
                promise.join()

    class context_manager:
        def calls_join_or_wait_on_close_of_block(self):
            promise = _runner().run(_, asynchronous=True)
            promise.join = Mock()
            with promise:
                pass
            promise.join.assert_called_once_with()

        def yields_self(self):
            promise = _runner().run(_, asynchronous=True)
            with promise as value:
                assert value is promise



from invoke.tasks import task


@task
def mytask(c):
    pass


@task
def basic_arg(c, arg="val"):
    pass


@task
def multiple_args(c, arg1="val1", otherarg="val2"):
    pass


@task
def basic_bool(c, mybool=True):
    pass



from queue import Queue

from invoke.util import ExceptionWrapper, ExceptionHandlingThread as EHThread


# TODO: rename
class ExceptionHandlingThread_:
    class via_target:
        def setup_method(self):
            def worker(q):
                q.put(7)

            self.worker = worker

        def base_case(self):
            queue = Queue()
            t = EHThread(target=self.worker, args=[queue])
            t.start()
            t.join()
            assert queue.get(block=False) == 7
            assert queue.empty()

        def catches_exceptions(self):
            # Induce exception by submitting a bad queue obj
            t = EHThread(target=self.worker, args=[None])
            t.start()
            t.join()
            wrapper = t.exception()
            assert isinstance(wrapper, ExceptionWrapper)
            assert wrapper.kwargs == {"args": [None], "target": self.worker}
            assert wrapper.type == AttributeError
            assert isinstance(wrapper.value, AttributeError)

        def exhibits_is_dead_flag(self):
            # Spin up a thread that will except internally (can't put() on a
            # None object)
            t = EHThread(target=self.worker, args=[None])
            t.start()
            t.join()
            # Excepted -> it's dead
            assert t.is_dead
            # Spin up a happy thread that can exit peacefully (it's not "dead",
            # though...maybe we should change that terminology)
            t = EHThread(target=self.worker, args=[Queue()])
            t.start()
            t.join()
            # Not dead, just uh...sleeping?
            assert not t.is_dead

    class via_subclassing:
        def setup_method(self):
            class MyThread(EHThread):
                def __init__(self, *args, **kwargs):
                    self.queue = kwargs.pop("queue")
                    super(MyThread, self).__init__(*args, **kwargs)

                def _run(self):
                    self.queue.put(7)

            self.klass = MyThread

        def base_case(self):
            queue = Queue()
            t = self.klass(queue=queue)
            t.start()
            t.join()
            assert queue.get(block=False) == 7
            assert queue.empty()

        def catches_exceptions(self):
            # Induce exception by submitting a bad queue obj
            t = self.klass(queue=None)
            t.start()
            t.join()
            wrapper = t.exception()
            assert isinstance(wrapper, ExceptionWrapper)
            assert wrapper.kwargs == {}
            assert wrapper.type == AttributeError
            assert isinstance(wrapper.value, AttributeError)

        def exhibits_is_dead_flag(self):
            # Spin up a thread that will except internally (can't put() on a
            # None object)
            t = self.klass(queue=None)
            t.start()
            t.join()
            # Excepted -> it's dead
            assert t.is_dead
            # Spin up a happy thread that can exit peacefully (it's not "dead",
            # though...maybe we should change that terminology)
            t = self.klass(queue=Queue())
            t.start()
            t.join()
            # Not dead, just uh...sleeping?
            assert not t.is_dead



from invoke.tasks import task
from invoke.collection import Collection


@task
def nope(c):
    return "You can't see this"


@task(autoprint=True)
def yup(c):
    return "It's alive!"


@task(pre=[yup])
def pre_check(c):
    pass


@task(post=[yup])
def post_check(c):
    pass


sub = Collection("sub", yup)
ns = Collection(nope, yup, pre_check, post_check, sub)
